#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import cherrypy
import json
import time

from ..describe import Description, autoDescribeRoute
from ..rest import Resource, setResponseHeader
from girder.api import access

# If no timeout param is passed to stream, we default to this value
DEFAULT_STREAM_TIMEOUT = 300
# When new events are seen, we will poll at the minimum interval
MIN_POLL_INTERVAL = 0.5
# The interval increases when no new events are seen, capping at this value
MAX_POLL_INTERVAL = 2


def sseMessage(event):
    """
    Serializes an event into the server-sent events protocol.
    """
    return 'data: %s\n\n' % json.dumps(event, default=str)


class Notification(Resource):
    def __init__(self):
        super(Notification, self).__init__()
        self.resourceName = 'notification'
        self.route('GET', ('stream',), self.stream)

    @access.cookie
    @access.token
    @autoDescribeRoute(
        Description('Stream notifications for a given user via the SSE protocol.')
        .notes('This uses long-polling to keep the connection open for '
               'several minutes at a time (or longer) and should be requested '
               'with an EventSource object or other SSE-capable client. '
               '<p>Notifications are returned within a few seconds of when '
               'they occur.  When no notification occurs for the timeout '
               'duration, the stream is closed. '
               '<p>This connection can stay open indefinitely long.')
        .param('timeout', 'The duration without a notification before the stream is closed.',
               dataType='integer', required=False, default=DEFAULT_STREAM_TIMEOUT)
        .errorResponse()
        .errorResponse('You are not logged in.', 403)
    )
    def stream(self, timeout, params):
        user, token = self.getCurrentUser(returnToken=True)

        setResponseHeader('Content-Type', 'text/event-stream')
        setResponseHeader('Cache-Control', 'no-cache')

        def streamGen():
            lastUpdate = None
            start = time.time()
            wait = MIN_POLL_INTERVAL
            while cherrypy.engine.state == cherrypy.engine.states.STARTED:
                wait = min(wait + MIN_POLL_INTERVAL, MAX_POLL_INTERVAL)
                for event in self.model('notification').get(
                        user, lastUpdate, token=token):
                    if lastUpdate is None or event['updated'] > lastUpdate:
                        lastUpdate = event['updated']
                    wait = MIN_POLL_INTERVAL
                    start = time.time()
                    yield sseMessage(event)
                if time.time() - start > timeout:
                    break

                time.sleep(wait)
        return streamGen



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import posixpath

from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, loadmodel, RestException
from girder.utility.progress import ProgressContext
from snakebite.client import Client as HdfsClient
from snakebite.errors import FileNotFoundException


class HdfsAssetstoreResource(Resource):
    def __init__(self):
        super(HdfsAssetstoreResource, self).__init__()
        self.resourceName = 'hdfs_assetstore'
        self.route('PUT', (':id', 'import'), self.importData)

        self.folderModel = self.model('folder')  # Save to avoid many lookups
        self.itemModel = self.model('item')
        self.fileModel = self.model('file')

    def _importFile(self, parent, name, user, assetstore, node):
        item = self.itemModel.findOne({
            'folderId': parent['_id'],
            'name': name
        })
        if item is None:
            item = self.itemModel.createItem(
                name=name, creator=user, folder=parent)

        file = self.fileModel.findOne({
            'name': name,
            'itemId': item['_id']
        })
        if file is None:
            file = self.fileModel.createFile(
                creator=user, item=item, name=name, size=node['length'],
                assetstore=assetstore, mimeType=None, saveFile=False)

        file['hdfs'] = {
            'imported': True,
            'path': node['path']
        }
        self.fileModel.save(file)

    def _importData(self, parentType, parent, assetstore, client, path, ctx,
                    user):
        for node in client.ls([path]):
            ctx.update(message='Importing ' + node['path'])
            name = posixpath.basename(node['path'])

            if node['file_type'] == 'd':
                folder = self.folderModel.findOne({
                    'parentId': parent['_id'],
                    'name': name,
                    'parentCollection': parentType
                })
                if folder is None:
                    folder = self.folderModel.createFolder(
                        parent, name, parentType=parentType, creator=user)

                self._importData('folder', folder, assetstore, client,
                                 node['path'], ctx, user)
            elif node['file_type'] == 'f' and parentType == 'folder':
                self._importFile(parent, name, user, assetstore, node)

    @access.admin
    @loadmodel(model='assetstore')
    @describeRoute(
        Description('Import a data hierarchy from an HDFS instance.')
        .notes('Only site administrators may use this endpoint.')
        .param('id', 'The ID of the assetstore representing the HDFS instance.',
               paramType='path')
        .param('parentId', 'The ID of the parent object in the Girder data '
               'hierarchy under which to import the files.')
        .param('parentType', 'The type of the parent object to import into.',
               enum=('folder', 'user', 'collection'), required=False)
        .param('path', 'Root of the directory structure (relative to the root '
               'of the HDFS) to import.')
        .param('progress', 'Whether to record progress on this operation ('
               'default=False)', required=False, dataType='boolean')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def importData(self, assetstore, params):
        self.requireParams(('parentId', 'path'), params)

        user = self.getCurrentUser()
        parentType = params.get('parentType', 'folder')
        if parentType not in ('user', 'collection', 'folder'):
            raise RestException('Invalid parentType.')

        parent = self.model(parentType).load(params['parentId'], force=True,
                                             exc=True)

        progress = self.boolParam('progress', params, default=False)
        client = HdfsClient(
            host=assetstore['hdfs']['host'], port=assetstore['hdfs']['port'],
            use_trash=False)
        path = params['path']

        with ProgressContext(progress, user=user,
                             title='Importing data from HDFS') as ctx:
            try:
                self._importData(
                    parentType, parent, assetstore, client, path, ctx, user)
            except FileNotFoundException:
                raise RestException('File not found: %s.' % path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2016 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, loadmodel, RestException
from girder.constants import AccessType, TokenScope
from girder.utility import mail_utils
from girder.utility.progress import ProgressContext
import datetime


CURATION = 'curation'

ENABLED = 'enabled'
STATUS = 'status'
TIMELINE = 'timeline'
ENABLE_USER_ID = 'enableUserId'
REQUEST_USER_ID = 'requestUserId'
REVIEW_USER_ID = 'reviewUserId'

CONSTRUCTION = 'construction'
REQUESTED = 'requested'
APPROVED = 'approved'

DEFAULTS = {
    ENABLED: False,
    STATUS: CONSTRUCTION,
}

TITLE_PUBLIC = 'Making curated folder public...'
TITLE_PRIVATE = 'Making curated folder private...'
TITLE_WRITEABLE = 'Making curated folder writeable...'
TITLE_READ_ONLY = 'Making curated folder read-only...'


class CuratedFolder(Resource):

    @access.user(scope=TokenScope.DATA_READ)
    @loadmodel(model='folder', level=AccessType.READ)
    @describeRoute(
        Description('Get curation details for the folder.')
        .param('id', 'The folder ID', paramType='path')
        .errorResponse('ID was invalid.')
        .errorResponse('Read permission denied on the folder.', 403)
    )
    def getCuration(self, folder, params):
        result = dict(DEFAULTS)
        result[TIMELINE] = []
        result.update(folder.get(CURATION, {}))
        result['public'] = folder.get('public')
        return result

    @access.user(scope=TokenScope.DATA_WRITE)
    @loadmodel(model='folder', level=AccessType.WRITE)
    @describeRoute(
        Description('Set curation details for the folder.')
        .param('id', 'The folder ID', paramType='path')
        .param('enabled', 'Enable or disable folder curation.',
               required=False, dataType='boolean')
        .param('status', 'Set the folder curation status.',
               required=False, dataType='string',
               enum=[CONSTRUCTION, REQUESTED, APPROVED])
        .errorResponse('ID was invalid.')
        .errorResponse('Write permission denied on the folder.', 403)
    )
    def setCuration(self, folder, params):
        user = self.getCurrentUser()
        if CURATION not in folder:
            folder[CURATION] = dict(DEFAULTS)
        curation = folder[CURATION]
        oldCuration = dict(curation)

        # update enabled
        oldEnabled = curation.get(ENABLED, False)
        if ENABLED in params:
            self.requireAdmin(user)
            curation[ENABLED] = self.boolParam(ENABLED, params)
        enabled = curation.get(ENABLED, False)

        # update status
        oldStatus = curation.get(STATUS)
        if STATUS in params:
            # verify valid status
            if params[STATUS] not in [CONSTRUCTION, REQUESTED, APPROVED]:
                raise RestException('Invalid status parameter')
            # regular users can only do construction -> requested transition
            if curation.get(STATUS) != CONSTRUCTION or \
               params[STATUS] != REQUESTED:
                self.requireAdmin(user)
            curation[STATUS] = params[STATUS]
        status = curation.get(STATUS)

        # admin enabling curation
        if enabled and not oldEnabled:
            # TODO: check if writers exist?
            # TODO: email writers?
            with self._progressContext(folder, TITLE_PRIVATE) as pc:
                self._setPublic(folder, False, pc)
            curation[ENABLE_USER_ID] = user.get('_id')
            self._addTimeline(oldCuration, curation, 'enabled curation')

        # admin reopening folder
        if enabled and oldStatus == APPROVED and status == CONSTRUCTION:
            with self._progressContext(folder, TITLE_PRIVATE) as pc:
                self._setPublic(folder, False, pc)
            curation[ENABLE_USER_ID] = user.get('_id')
            with self._progressContext(folder, TITLE_WRITEABLE) as pc:
                self._makeWriteable(folder, pc)
            self._addTimeline(oldCuration, curation, 'reopened folder')

        # admin disabling curation
        if not enabled and oldEnabled:
            self._addTimeline(oldCuration, curation, 'disabled curation')

        # user requesting approval
        if enabled and oldStatus == CONSTRUCTION and status == REQUESTED:
            curation[REQUEST_USER_ID] = user.get('_id')
            with self._progressContext(folder, TITLE_READ_ONLY) as pc:
                self._makeReadOnly(folder, pc)
            self._addTimeline(oldCuration, curation, 'requested approval')
            # send email to admin requesting approval
            self._sendMail(
                folder, curation.get(ENABLE_USER_ID),
                'REQUEST FOR APPROVAL: ' + folder['name'],
                'curation.requested.mako')

        # admin approving request
        if enabled and oldStatus == REQUESTED and status == APPROVED:
            with self._progressContext(folder, TITLE_PUBLIC) as pc:
                self._setPublic(folder, True, pc)
            curation[REVIEW_USER_ID] = user.get('_id')
            self._addTimeline(oldCuration, curation, 'approved request')
            # send approval notification to requestor
            self._sendMail(
                folder, curation.get(REQUEST_USER_ID),
                'APPROVED: ' + folder['name'],
                'curation.approved.mako')

        # admin rejecting request
        if enabled and oldStatus == REQUESTED and status == CONSTRUCTION:
            curation[REVIEW_USER_ID] = user.get('_id')
            with self._progressContext(folder, TITLE_WRITEABLE) as pc:
                self._makeWriteable(folder, pc)
            self._addTimeline(oldCuration, curation, 'rejected request')
            # send rejection notification to requestor
            self._sendMail(
                folder, curation.get(REQUEST_USER_ID),
                'REJECTED: ' + folder['name'],
                'curation.rejected.mako')

        folder = self.model('folder').save(folder)
        curation['public'] = folder.get('public')
        return curation

    def _progressContext(self, folder, title):
        user = self.getCurrentUser()
        total = self.model('folder').subtreeCount(folder, includeItems=False)
        return ProgressContext(True, user=user, total=total, title=title)

    def _setPublic(self, folder, public, pc):
        """
        Recursively updates folder's public setting.
        """
        pc.update(increment=1)
        folder['public'] = public
        folder = self.model('folder').save(folder)
        subfolders = self.model('folder').find({
            'parentId': folder['_id'],
            'parentCollection': 'folder'
        })
        for folder in subfolders:
            self._setPublic(folder, public, pc)

    def _makeReadOnly(self, folder, pc):
        """
        Recursively updates folder permissions so that anyone with write access
        now has read-only access.
        """
        pc.update(increment=1)
        for doc in folder['access']['users'] + folder['access']['groups']:
            if doc['level'] == AccessType.WRITE:
                doc['level'] = AccessType.READ
        folder = self.model('folder').save(folder)
        subfolders = self.model('folder').find({
            'parentId': folder['_id'],
            'parentCollection': 'folder'
        })
        for folder in subfolders:
            self._makeReadOnly(folder, pc)

    def _makeWriteable(self, folder, pc):
        """
        Recursively updates folder permissions so that anyone with read access
        now has write access.
        """
        pc.update(increment=1)
        for doc in folder['access']['users'] + folder['access']['groups']:
            if doc['level'] == AccessType.READ:
                doc['level'] = AccessType.WRITE
        folder = self.model('folder').save(folder)
        subfolders = self.model('folder').find({
            'parentId': folder['_id'],
            'parentCollection': 'folder'
        })
        for folder in subfolders:
            self._makeWriteable(folder, pc)

    def _addTimeline(self, oldCuration, curation, text):
        """
        Adds a new entry to the curation timeline.

        :param oldCuration: the curation values before the last change
        :param curation: the curation values after the last change
        :param text: a short human-readable description of the change
        """
        user = self.getCurrentUser()
        data = dict(
            userId=user.get('_id'),
            userLogin=user.get('login'),
            text=text,
            oldEnabled=oldCuration[ENABLED],
            oldStatus=oldCuration[STATUS],
            enabled=curation[ENABLED],
            status=curation[STATUS],
            timestamp=datetime.datetime.utcnow())
        curation.setdefault(TIMELINE, []).append(data)

    def _getEmail(self, userId):
        """
        Loads and returns the email address for the specified user.
        """
        return self.model('user').load(userId, force=True).get('email')

    def _sendMail(self, folder, userId, subject, template):
        """
        Sends the specified email template to a single user.

        :param folder: the curated folder
        :param userId: the id of the user to email
        :param subject: the email subject
        :param template: the name of the mako template to use
        """
        if not userId:
            return
        data = dict(
            folder=folder,
            curation=folder[CURATION])
        text = mail_utils.renderTemplate(template, data)
        emails = [self._getEmail(userId)]
        mail_utils.sendEmail(emails, subject, text)


def load(info):
    curatedFolder = CuratedFolder()
    info['apiRoot'].folder.route(
        'GET', (':id', 'curation'), curatedFolder.getCuration)
    info['apiRoot'].folder.route(
        'PUT', (':id', 'curation'), curatedFolder.setCuration)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import cherrypy  # pragma: no cover
import argparse  # pragma: no cover
import os  # pragma: no cover

try:  # pragma: no cover
    from girder.utility import server
except ImportError:
    # Update python path to ensure server respawning works. See #732
    source_root_dir = os.path.dirname(os.path.dirname(__file__))
    import sys
    cherrypy.engine.log("[Girder] Appending source root dir to 'sys.path': %s"
                        % source_root_dir)
    sys.path.append(source_root_dir)
    from girder.utility import server


def main():
    parser = argparse.ArgumentParser(
        description='Girder: data management platform for the web.')
    parser.add_argument("-t", "--testing", help="run in testing mode",
                        action="store_true")
    parser.add_argument("-d", "--database",
                        help="to what database url should Girder connect")
    parser.add_argument("-p", "--port",
                        help="on what port should Girder serve")
    args = parser.parse_args()
    if args.database:
        cherrypy.config['database']['uri'] = args.database
    if args.port:
        cherrypy.config['server.socket_port'] = int(args.port)
    server.setup(args.testing)

    cherrypy.engine.start()
    cherrypy.engine.block()


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from . geospatial import GeospatialItem


def load(info):
    geospatialItem = GeospatialItem()
    info['apiRoot'].item.route('POST', ('geospatial',),
                               geospatialItem.create)
    info['apiRoot'].item.route('GET', ('geospatial',),
                               geospatialItem.find)
    info['apiRoot'].item.route('GET', ('geospatial', 'intersects'),
                               geospatialItem.intersects)
    info['apiRoot'].item.route('GET', ('geospatial', 'near'),
                               geospatialItem.near)
    info['apiRoot'].item.route('GET', ('geospatial', 'within'),
                               geospatialItem.within)
    info['apiRoot'].item.route('GET', (':id', 'geospatial'),
                               geospatialItem.getGeospatial)
    info['apiRoot'].item.route('PUT', (':id', 'geospatial'),
                               geospatialItem.setGeospatial)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from .assetstore import HdfsAssetstoreAdapter
from .rest import HdfsAssetstoreResource
from girder import events
from girder.api import access
from girder.api.v1.assetstore import Assetstore
from girder.constants import AssetstoreType
from girder.utility.model_importer import ModelImporter
from girder.utility import assetstore_utilities


def getAssetstore(event):
    assetstore = event.info
    if assetstore['type'] == AssetstoreType.HDFS:
        event.stopPropagation()
        event.addResponse(HdfsAssetstoreAdapter)


def updateAssetstore(event):
    params = event.info['params']
    assetstore = event.info['assetstore']

    if assetstore['type'] == AssetstoreType.HDFS:
        assetstore['hdfs'] = {
            'host': params.get('hdfsHost', assetstore['hdfs']['host']),
            'port': params.get('hdfsPort', assetstore['hdfs']['port']),
            'path': params.get('hdfsPath', assetstore['hdfs']['path']),
            'webHdfsPort': params.get('webHdfsPort',
                                      assetstore['hdfs'].get('webHdfsPort')),
            'user': params.get('hdfsUser', assetstore['hdfs'].get('user'))
        }


@access.admin
def createAssetstore(event):
    params = event.info['params']

    if params.get('type') == AssetstoreType.HDFS:
        event.addResponse(ModelImporter.model('assetstore').save({
            'type': AssetstoreType.HDFS,
            'name': params.get('name'),
            'hdfs': {
                'host': params.get('host'),
                'port': params.get('port'),
                'path': params.get('path'),
                'webHdfsPort': params.get('webHdfsPort'),
                'user': params.get('effectiveUser')
            }
        }))
        event.preventDefault()


def load(info):
    AssetstoreType.HDFS = 'hdfs'
    events.bind('assetstore.update', 'hdfs_assetstore', updateAssetstore)
    events.bind('rest.post.assetstore.before', 'hdfs_assetstore',
                createAssetstore)

    assetstore_utilities.setAssetstoreAdapter(AssetstoreType.HDFS, HdfsAssetstoreAdapter)

    (Assetstore.createAssetstore.description
        .param('host', 'The namenode host (for HDFS type).', required=False)
        .param('port', 'The namenode RPC port (for HDFS type).', required=False)
        .param('path', 'Absolute path under which new files will be stored ('
               'for HDFS type).', required=False)
        .param('user', 'The effective user to use when calling HDFS RPCs (for '
               'HDFS type). This defaults to whatever system username the '
               'Girder server process is running under.', required=False)
        .param('webHdfsPort', 'WebHDFS port for the namenode. You must enable '
               'WebHDFS on your Hadoop cluster if you want to write new files '
               'to the assetstore (for HDFS type).', required=False))

    info['apiRoot'].hdfs_assetstore = HdfsAssetstoreResource()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource, filtermodel
from girder.constants import AccessType, SortDir


class Job(Resource):
    def __init__(self):
        super(Job, self).__init__()
        self.resourceName = 'job'

        self.route('GET', (), self.listJobs)
        self.route('GET', (':id',), self.getJob)
        self.route('PUT', (':id',), self.updateJob)
        self.route('DELETE', (':id',), self.deleteJob)

    @access.public
    @filtermodel(model='job', plugin='jobs')
    @autoDescribeRoute(
        Description('List jobs for a given user.')
        .param('userId', 'The ID of the user whose jobs will be listed. If '
               'not passed or empty, will use the currently logged in user. If '
               'set to "None", will list all jobs that do not have an owning '
               'user.', required=False)
        .pagingParams(defaultSort='created', defaultSortDir=SortDir.DESCENDING)
    )
    def listJobs(self, userId, limit, offset, sort, params):
        currentUser = self.getCurrentUser()
        if not userId:
            user = currentUser
        elif userId.lower() == 'none':
            user = None
        else:
            user = self.model('user').load(userId, user=currentUser, level=AccessType.READ)

        return list(self.model('job', 'jobs').list(
            user=user, offset=offset, limit=limit, sort=sort, currentUser=currentUser))

    @access.public
    @filtermodel(model='job', plugin='jobs')
    @autoDescribeRoute(
        Description('Get a job by ID.')
        .modelParam('id', 'The ID of the job.', model='job', plugin='jobs', level=AccessType.READ,
                    includeLog=True)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the job.', 403)
    )
    def getJob(self, job, params):
        return job

    @access.token
    @filtermodel(model='job', plugin='jobs')
    @autoDescribeRoute(
        Description('Update an existing job.')
        .notes('In most cases, regular users should not call this endpoint. It '
               'will typically be used by a batch processing system to send '
               'updates regarding the execution of the job. If using a non-'
               'user-associated token for authorization, the token must be '
               'granted the "jobs.job_<id>" scope, where <id> is the ID of '
               'the job being updated.')
        .modelParam('id', 'The ID of the job.', model='job', plugin='jobs', force=True)
        .param('log', 'A message to add to the job\'s log field. If you want '
               'to overwrite any existing log content, pass another parameter '
               '"overwrite=true".', required=False)
        .param('overwrite', 'If passing a log parameter, you may set this to '
               '"true" if you wish to overwrite the log field rather than '
               'append to it.', dataType='boolean', required=False, default=False)
        .param('status', 'Update the status of the job. See the JobStatus '
               'enumeration in the constants module in this plugin for the '
               'numerical values of each status.', required=False)
        .param('progressTotal', 'Maximum progress value, set <= 0 to indicate '
               'indeterminate progress for this job.', required=False, dataType='float')
        .param('progressCurrent', 'Current progress value.', required=False, dataType='float')
        .param('progressMessage', 'Current progress message.', required=False)
        .param('notify', 'If this update should trigger a notification, set '
               'this field to true.', dataType='boolean', required=False, default=True)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the job.', 403)
    )
    def updateJob(self, job, log, overwrite, notify, status, progressTotal, progressCurrent,
                  progressMessage, params):
        user = self.getCurrentUser()
        if user:
            self.model('job', 'jobs').requireAccess(job, user, level=AccessType.WRITE)
        else:
            self.ensureTokenScopes('jobs.job_' + str(job['_id']))

        return self.model('job', 'jobs').updateJob(
            job, log=log, status=status, overwrite=overwrite, notify=notify,
            progressCurrent=progressCurrent, progressTotal=progressTotal,
            progressMessage=progressMessage)

    @access.user
    @autoDescribeRoute(
        Description('Delete an existing job.')
        .modelParam('id', 'The ID of the job.', model='job', plugin='jobs', level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the job.', 403)
    )
    def deleteJob(self, job, params):
        self.model('job', 'jobs').remove(job)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import six

from geojson import GeoJSON
from pymongo import GEOSPHERE
from pymongo.errors import OperationFailure

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource, RestException
from girder.constants import AccessType


GEOSPATIAL_FIELD = 'geo'


class GeospatialItem(Resource):
    """
    Geospatial methods added to the API endpoint for items.
    """

    @access.user
    @autoDescribeRoute(
        Description('Create new items from a GeoJSON feature or feature collection.')
        .modelParam('folderId', 'The ID of the parent folder.', model='folder',
                    level=AccessType.WRITE, paramType='formData')
        .jsonParam('geoJSON', 'A GeoJSON object containing the features or feature'
                   ' collection to add.')
        .errorResponse()
        .errorResponse('Invalid GeoJSON was passed in request body.')
        .errorResponse('GeoJSON feature or feature collection was not passed in'
                       ' request body.')
        .errorResponse("GeoJSON feature did not contain a property named"
                       " 'name'.")
        .errorResponse('Property name was invalid.')
        .errorResponse('Write access was denied on the parent folder.', 403)
        .notes("All GeoJSON features must contain a property named 'name' from"
               " which the name of each created item is taken.")
    )
    def create(self, folder, geoJSON, params):
        try:
            GeoJSON.to_instance(geoJSON, strict=True)
        except ValueError:
            raise RestException('Invalid GeoJSON passed in request body.')

        if geoJSON['type'] == 'Feature':
            features = [geoJSON]
        elif geoJSON['type'] == 'FeatureCollection':
            features = geoJSON['features']
        else:
            raise RestException('GeoJSON feature or feature collection must be '
                                'passed in request body.')

        data = []

        for feature in features:
            properties = feature['properties']
            if 'name' not in properties:
                raise RestException("All GeoJSON features must contain a"
                                    " property named 'name'.")
            name = properties['name']
            del properties['name']
            if 'description' in properties:
                description = properties['description']
                del properties['description']
            else:
                description = ''

            for key in properties:
                if not len(key):
                    raise RestException('Property names must be at least one'
                                        ' character long.')
                if '.' in key or key[0] == '$':
                    raise RestException('The property name %s must not contain'
                                        ' a period or begin with a dollar sign.' % key)

            data.append({'name': name,
                         'description': description,
                         'metadata': properties,
                         'geometry': feature['geometry']})

        user = self.getCurrentUser()

        items = []

        for datum in data:
            newItem = self.model('item').createItem(
                folder=folder, name=datum['name'], creator=user,
                description=datum['description'])
            self.model('item').setMetadata(newItem, datum['metadata'])
            newItem[GEOSPATIAL_FIELD] = {'geometry': datum['geometry']}
            newItem = self.model('item').updateItem(newItem)
            items.append(newItem)

        return [self._filter(item) for item in items]

    @access.public
    @autoDescribeRoute(
        Description('Search for an item by geospatial data.')
        .jsonParam('q', 'Search query as a JSON object.')
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
    )
    def find(self, q, limit, offset, sort, params):
        return self._find(q, limit, offset, sort)

    @access.public
    @autoDescribeRoute(
        Description('Search for items that intersects with a GeoJSON object.')
        .param('field', 'Name of field containing GeoJSON on which to search.', strip=True)
        .jsonParam('geometry', 'Search query condition as a GeoJSON object.')
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
    )
    def intersects(self, field, geometry, limit, offset, sort, params):
        try:
            GeoJSON.to_instance(geometry, strict=True)
        except (TypeError, ValueError):
            raise RestException("Invalid GeoJSON passed as 'geometry' parameter.")

        if field[:3] != '%s.' % GEOSPATIAL_FIELD:
            field = '%s.%s' % (GEOSPATIAL_FIELD, field)

        query = {
            field: {
                '$geoIntersects': {
                    '$geometry': geometry
                }
            }
        }

        return self._find(query, limit, offset, sort)

    def _getGeometry(self, geometry):
        try:
            GeoJSON.to_instance(geometry, strict=True)

            if geometry['type'] != 'Point':
                raise ValueError

            return geometry
        except (TypeError, ValueError):
            raise RestException("Invalid GeoJSON passed as 'geometry' parameter.")

    @access.public
    @autoDescribeRoute(
        Description('Search for items that are in proximity to a GeoJSON point.')
        .param('field', 'Name of field containing GeoJSON on which to search.', strip=True)
        .jsonParam('geometry', 'Search query condition as a GeoJSON point.')
        .param('maxDistance', 'Limits results to items that are at most this distance '
               'in meters from the GeoJSON point.', required=False, dataType='number')
        .param('minDistance', 'Limits results to items that are at least this distance '
               'in meters from the GeoJSON point.', required=False, dataType='number')
        .param('ensureIndex', 'Create a 2dsphere index on the field on which to search '
               'if one does not exist.', required=False, dataType='boolean', default=False)
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
        .errorResponse('Field on which to search was not indexed.')
        .errorResponse('Index creation was denied.', 403)
        .notes("Field on which to search be indexed by a 2dsphere index."
               " Anonymous users may not use 'ensureIndex' to create such an index.")
    )
    def near(self, field, geometry, maxDistance, minDistance, ensureIndex, limit,
             offset, sort, params):
        condition = {
            '$geometry': self._getGeometry(geometry)
        }

        if maxDistance is not None:
            if maxDistance < 0:
                raise RestException('maxDistance must be positive.')
            condition['$maxDistance'] = maxDistance
        if minDistance is not None:
            if minDistance < 0:
                raise RestException('minDistance must be positive.')
            condition['$minDistance'] = minDistance

        if field[:3] != '%s.' % GEOSPATIAL_FIELD:
            field = '%s.%s' % (GEOSPATIAL_FIELD, field)

        if ensureIndex:
            user = self.getCurrentUser()

            if not user:
                raise RestException('Index creation denied.', 403)

            self.model('item').collection.create_index([(field, GEOSPHERE)])

        query = {
            field: {
                '$near': condition
            }
        }

        try:
            return self._find(query, limit, offset, sort)
        except OperationFailure:
            raise RestException("Field '%s' must be indexed by a 2dsphere index." % field)

    _RADIUS_OF_EARTH = 6378137.0  # average in meters

    @access.public
    @autoDescribeRoute(
        Description('Search for items that are entirely within either a GeoJSON'
                    ' polygon or a circular region.')
        .param('field', 'Name of field containing GeoJSON on which to search.', strip=True)
        .jsonParam('geometry', 'Search query condition as a GeoJSON polygon.',
                   required=False)
        .jsonParam('center', 'Center of search radius as a GeoJSON point.',
                   required=False, requireObject=True)
        .param('radius', 'Search radius in meters.', required=False,
               dataType='number')
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
        .errorResponse('Field on which to search was not indexed.')
        .errorResponse('Index creation was denied.', 403)
        .notes("Either parameter 'geometry' or both parameters 'center' "
               " and 'radius' are required.")
    )
    def within(self, field, geometry, center, radius, limit, offset, sort, params):
        if geometry is not None:
            try:
                GeoJSON.to_instance(geometry, strict=True)

                if geometry['type'] != 'Polygon':
                    raise ValueError
            except (TypeError, ValueError):
                raise RestException("Invalid GeoJSON passed as 'geometry' parameter.")

            condition = {
                '$geometry': geometry
            }

        elif center is not None and radius is not None:
            try:
                radius /= self._RADIUS_OF_EARTH

                if radius < 0.0:
                    raise ValueError
            except ValueError:
                raise RestException("Parameter 'radius' must be a number.")

            try:
                GeoJSON.to_instance(center, strict=True)

                if center['type'] != 'Point':
                    raise ValueError
            except (TypeError, ValueError):
                raise RestException("Invalid GeoJSON passed as 'center' parameter.")

            condition = {
                '$centerSphere': [center['coordinates'], radius]
            }

        else:
            raise RestException("Either parameter 'geometry' or both parameters"
                                " 'center' and 'radius' are required.")

        if field[:3] != '%s.' % GEOSPATIAL_FIELD:
            field = '%s.%s' % (GEOSPATIAL_FIELD, field)

        query = {
            field: {
                '$geoWithin': condition
            }
        }

        return self._find(query, limit, offset, sort)

    @access.public
    @autoDescribeRoute(
        Description('Get an item and its geospatial data by ID.')
        .modelParam('id', 'The ID of the item.', model='item', level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the item.', 403)
    )
    def getGeospatial(self, item, params):
        return self._filter(item)

    @access.user
    @autoDescribeRoute(
        Description('Set geospatial fields on an item.')
        .notes('Set geospatial fields to null to delete them.')
        .modelParam('id', 'The ID of the item.', model='item', level=AccessType.WRITE)
        .jsonParam('geospatial', 'A JSON object containing the geospatial fields to add.',
                   paramType='body')
        .errorResponse('ID was invalid.')
        .errorResponse('Invalid JSON was passed in request body.')
        .errorResponse('Geospatial key name was invalid.')
        .errorResponse('Geospatial field did not contain valid GeoJSON.')
        .errorResponse('Write access was denied for the item.', 403)
    )
    def setGeospatial(self, item, geospatial, params):
        for k, v in six.viewitems(geospatial):
            if '.' in k or k[0] == '$':
                raise RestException('Geospatial key name %s must not contain a'
                                    ' period or begin with a dollar sign.' % k)
            if v:
                try:
                    GeoJSON.to_instance(v, strict=True)
                except ValueError:
                    raise RestException('Geospatial field with key %s does not'
                                        ' contain valid GeoJSON: %s' % (k, v))

        if GEOSPATIAL_FIELD not in item:
            item[GEOSPATIAL_FIELD] = dict()

        item[GEOSPATIAL_FIELD].update(six.viewitems(geospatial))
        keys = [k for k, v in six.viewitems(item[GEOSPATIAL_FIELD]) if v is None]

        for key in keys:
            del item[GEOSPATIAL_FIELD][key]

        item = self.model('item').updateItem(item)

        return self._filter(item)

    def _filter(self, item):
        """
        Helper to filter the fields of an item and append its geospatial data.

        :param item: item whose fields to filter and geospatial data append.
        :type item: dict[str, unknown]
        :returns: filtered fields of the item with geospatial data appended to
                 its 'geo' field.
        :rtype : dict[str, unknown]
        """
        filtered = self.model('item').filter(item)

        if GEOSPATIAL_FIELD in item:
            filtered[GEOSPATIAL_FIELD] = item[GEOSPATIAL_FIELD]
        else:
            filtered[GEOSPATIAL_FIELD] = {}

        return filtered

    def _find(self, query, limit, offset, sort):
        """
        Helper to search the geospatial data of items and return the filtered
        fields and geospatial data of the matching items.

        :param query: geospatial search query.
        :type query: dict[str, unknown]
        :param limit: maximum number of matching items to return.
        :type limit: int
        :param offset: offset of matching items to return.
        :type offset: int
        :param sort: field by which to sort the matching items
        :type sort: str
        :returns: filtered fields of the matching items with geospatial data
                 appended to the 'geo' field of each item.
        :rtype : list[dict[str, unknown]]
        """
        user = self.getCurrentUser()
        cursor = self.model('item').find(query, sort=sort)

        return [self._filter(result) for result in
                self.model('item')
                    .filterResultsByPermission(cursor, user, AccessType.READ,
                                               limit, offset)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import struct
import jwt
from jwt.utils import base64url_decode
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
from six.moves import urllib
from six import string_types

from girder.api.rest import getApiUrl, RestException
from .base import ProviderBase
from .. import constants


class Globus(ProviderBase):
    _AUTH_URL = 'https://auth.globus.org/v2/oauth2/authorize'
    _AUTH_SCOPES = ('urn:globus:auth:scope:auth.globus.org:view_identities',
                    'openid', 'profile', 'email')
    _TOKEN_URL = 'https://auth.globus.org/v2/oauth2/token'
    _JWK_KEY_URL = 'https://auth.globus.org/jwk.json'
    jwk_keys = None

    def getClientIdSetting(self):
        return self.model('setting').get(
            constants.PluginSettings.GLOBUS_CLIENT_ID)

    def getClientSecretSetting(self):
        return self.model('setting').get(
            constants.PluginSettings.GLOBUS_CLIENT_SECRET)

    @classmethod
    def getUrl(cls, state):
        clientId = cls.model('setting').get(
            constants.PluginSettings.GLOBUS_CLIENT_ID)

        if clientId is None:
            raise Exception('No Globus client ID setting is present.')

        callbackUrl = '/'.join((getApiUrl(), 'oauth', 'globus', 'callback'))

        query = urllib.parse.urlencode({
            'response_type': 'code',
            'access_type': 'online',
            'client_id': clientId,
            'redirect_uri': callbackUrl,
            'state': state,
            'scope': ' '.join(cls._AUTH_SCOPES)
        })
        return '%s?%s' % (cls._AUTH_URL, query)

    def getToken(self, code):
        params = {
            'grant_type': 'authorization_code',
            'code': code,
            'client_id': self.clientId,
            'client_secret': self.clientSecret,
            'redirect_uri': self.redirectUri
        }
        resp = self._getJson(method='POST', url=self._TOKEN_URL,
                             data=params,
                             headers={'Accept': 'application/json'})
        if 'error' in resp:
            raise RestException(
                'Got an error exchanging token from provider: "%s".' % resp,
                code=502)
        return resp

    def getUser(self, token):
        id_token = token.get('id_token')

        if self.jwk_keys is None:
            # Get public keys required for decoding 'id_token'
            self.jwk_keys = self._getJson(method='GET', url=self._JWK_KEY_URL)

        # There should be only one entry
        keyobj = self.jwk_keys['keys'][0]

        # Borrowed from 'cryptography' for py2 backward compat
        def decode_value(val):
            if hasattr(int, 'from_bytes'):
                int_from_bytes = int.from_bytes
            else:
                def int_from_bytes(data, byteorder, signed=False):
                    assert byteorder == 'big'
                    assert not signed

                    if len(data) % 4 != 0:
                        data = (b'\x00' * (4 - (len(data) % 4))) + data

                    result = 0

                    while len(data) > 0:
                        digit, = struct.unpack('>I', data[:4])
                        result = (result << 32) + digit
                        data = data[4:]
                    return result

            if isinstance(val, string_types):
                val = val.encode('utf-8')
            decoded = base64url_decode(val)
            return int_from_bytes(decoded, 'big')

        # Create RSA key from JSON spec
        key = rsa.RSAPublicNumbers(
            n=decode_value(keyobj['n']),
            e=decode_value(keyobj['e'])
        ).public_key(default_backend())

        # Decode 'id_token'
        # BEWARE: leeway should be 0, but upstream server seems to be desynch
        identity = jwt.decode(
            id_token, key, algorithm='RS512', leeway=100,
            audience=self.clientId)

        oauthId = identity.get('sub')
        if not oauthId:
            raise RestException(
                'Globus identity did not return a valid ID.', code=502)

        email = identity.get('email')
        if not email:
            raise RestException(
                'Globus identity did not return a valid email.', code=502)

        name = identity['name'].split()
        firstName = name[0]
        lastName = name[-1]

        user = self._createOrReuseUser(oauthId, email, firstName, lastName)
        return user



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder.api import access
from girder.api.describe import autoDescribeRoute, Description
from girder.api.rest import RestException, setRawResponse, setResponseHeader
from girder.api.v1.file import File
from girder.constants import AccessType, TokenScope
from girder.utility.model_importer import ModelImporter


class HashedFile(File):

    supportedAlgorithms = {
        'sha512'
    }

    def __init__(self, node):
        super(File, self).__init__()

        node.route('GET', ('hashsum', ':algo', ':hash', 'download'),
                   self.downloadWithHash)
        node.route('GET', (':id', 'hashsum_file', ':algo'),
                   self.downloadKeyFile)

    @access.cookie
    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Download the hashsum key file for a given file.')
        .modelParam('id', 'The ID of the file.', model='file', level=AccessType.READ)
        .param('algo', 'The hashsum algorithm.', paramType='path', lower=True,
               enum=supportedAlgorithms)
        .notes('This is meant to be used in conjunction with CMake\'s ExternalData module.')
        .errorResponse()
        .errorResponse('Read access was denied on the file.', 403)
    )
    def downloadKeyFile(self, file, algo, params):
        self._validateAlgo(algo)

        if algo not in file:
            raise RestException('This file does not have the %s hash computed.' % algo)
        hash = file[algo]
        name = '.'.join((file['name'], algo))

        setResponseHeader('Content-Length', len(hash))
        setResponseHeader('Content-Type', 'text/plain')
        setResponseHeader('Content-Disposition', 'attachment; filename="%s"' % name)
        setRawResponse()

        return hash

    @access.cookie
    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Download a file by its hash sum.')
        .param('algo', 'The type of the given hash sum (case insensitive).',
               paramType='path', lower=True, enum=supportedAlgorithms)
        .param('hash', 'The hexadecimal hash sum of the file to download (case insensitive).',
               paramType='path', lower=True)
        .errorResponse('No file with the given hash exists.')
    )
    def downloadWithHash(self, algo, hash, params):
        file = self._getFirstFileByHash(algo, hash)
        if not file:
            raise RestException('File not found.', code=404)

        return self.download(id=file['_id'], params=params)

    def _validateAlgo(self, algo):
        """
        Print an exception if a user requests an invalid checksum algorithm.
        """
        if algo not in self.supportedAlgorithms:
            msg = 'Invalid algorithm "%s". Supported algorithms: %s.' % (
                algo, ', '.join(self.supportedAlgorithms))
            raise RestException(msg, code=400)

    def _getFirstFileByHash(self, algo, hash, user=None):
        """
        Return the first file that the user has access to given its hash and its
        associated hash sum algorithm name.

        :param algo: Algorithm the given hash is encoded with.
        :param hash: Hash of the file to find.
        :param user: User to test access against.
         Default (none) is the current user.
        :return: A file document.
        """
        self._validateAlgo(algo)

        query = {algo: hash}  # Always convert to lower case
        fileModel = self.model('file')
        cursor = fileModel.find(query)

        if not user:
            user = self.getCurrentUser()

        for file in cursor:
            if fileModel.hasAccess(file, user, AccessType.READ):
                return file

        return None


def load(info):
    HashedFile(info['apiRoot'].file)
    ModelImporter.model('file').exposeFields(
        level=AccessType.READ, fields=HashedFile.supportedAlgorithms)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from ..describe import Description, autoDescribeRoute
from ..rest import Resource, RestException
from girder import events
from girder.constants import AccessType, AssetstoreType, TokenScope
from girder.api import access
from girder.utility.progress import ProgressContext


class Assetstore(Resource):
    """
    API Endpoint for managing assetstores. Requires admin privileges.
    """
    def __init__(self):
        super(Assetstore, self).__init__()
        self.resourceName = 'assetstore'
        self.route('GET', (), self.find)
        self.route('GET', (':id',), self.getAssetstore)
        self.route('POST', (), self.createAssetstore)
        self.route('POST', (':id', 'import'), self.importData)
        self.route('PUT', (':id',), self.updateAssetstore)
        self.route('DELETE', (':id',), self.deleteAssetstore)
        self.route('GET', (':id', 'files'), self.getAssetstoreFiles)

    @access.admin
    @autoDescribeRoute(
        Description('Get information about an assetstore.')
        .modelParam('id', model='assetstore')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def getAssetstore(self, assetstore, params):
        self.model('assetstore').addComputedInfo(assetstore)
        return assetstore

    @access.admin
    @autoDescribeRoute(
        Description('List assetstores.')
        .pagingParams(defaultSort='name')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def find(self, limit, offset, sort, params):
        return list(self.model('assetstore').list(offset=offset, limit=limit, sort=sort))

    @access.admin
    @autoDescribeRoute(
        Description('Create a new assetstore.')
        .responseClass('Assetstore')
        .notes('You must be an administrator to call this.')
        .param('name', 'Unique name for the assetstore.')
        .param('type', 'Type of the assetstore.', dataType='integer')
        .param('root', 'Root path on disk (for filesystem type).',
               required=False)
        .param('perms', 'File creation permissions (for filesystem type).',
               required=False)
        .param('db', 'Database name (for GridFS type)', required=False)
        .param('mongohost', 'Mongo host URI (for GridFS type)', required=False)
        .param('replicaset', 'Replica set name (for GridFS type)',
               required=False)
        .param('bucket', 'The S3 bucket to store data in (for S3 type).',
               required=False)
        .param('prefix', 'Optional path prefix within the bucket under which '
               'files will be stored (for S3 type).', required=False, default='')
        .param('accessKeyId', 'The AWS access key ID to use for authentication '
               '(for S3 type).', required=False)
        .param('secret', 'The AWS secret key to use for authentication (for '
               'S3 type).', required=False)
        .param('service', 'The S3 service host (for S3 type).  Default is '
               's3.amazonaws.com.  This can be used to specify a protocol and '
               'port as well using the form [http[s]://](host domain)[:(port)]. '
               'Do not include the bucket name here.', required=False, default='')
        .param('readOnly', 'If this assetstore is read-only, set this to true.',
               required=False, dataType='boolean', default=False)
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def createAssetstore(self, name, type, root, perms, db, mongohost, replicaset, bucket,
                         prefix, accessKeyId, secret, service, readOnly, params):
        if type == AssetstoreType.FILESYSTEM:
            self.requireParams({'root': root})
            return self.model('assetstore').createFilesystemAssetstore(
                name=name, root=root, perms=perms)
        elif type == AssetstoreType.GRIDFS:
            self.requireParams({'db': db})
            return self.model('assetstore').createGridFsAssetstore(
                name=name, db=db, mongohost=mongohost, replicaset=replicaset)
        elif type == AssetstoreType.S3:
            self.requireParams({'bucket': bucket})
            return self.model('assetstore').createS3Assetstore(
                name=name, bucket=bucket, prefix=prefix, secret=secret,
                accessKeyId=accessKeyId, service=service, readOnly=readOnly)
        else:
            raise RestException('Invalid type parameter')

    @access.admin(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Import existing data into an assetstore.')
        .notes('This does not move or copy the existing data, it just creates '
               'references to it in the Girder data hierarchy. Deleting '
               'those references will not delete the underlying data. This '
               'operation is currently only supported for S3 assetstores.')
        .modelParam('id', model='assetstore')
        .param('importPath', 'Root path within the underlying storage system '
               'to import.', required=False)
        .param('destinationId', 'ID of a folder, collection, or user in Girder '
               'under which the data will be imported.')
        .param('destinationType', 'Type of the destination resource.',
               enum=('folder', 'collection', 'user'))
        .param('progress', 'Whether to record progress on the import.',
               dataType='boolean', default=False, required=False)
        .param('leafFoldersAsItems', 'Whether folders containing only files should be '
               'imported as items.', dataType='boolean', required=False, default=False)
        .param('fileIncludeRegex', 'If set, only filenames matching this regular '
               'expression will be imported.', required=False)
        .param('fileExcludeRegex', 'If set, only filenames that do not match this regular '
               'expression will be imported. If a file matches both the include and exclude regex, '
               'it will be excluded.', required=False)
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def importData(self, assetstore, importPath, destinationId, destinationType, progress,
                   leafFoldersAsItems, fileIncludeRegex, fileExcludeRegex, params):
        user = self.getCurrentUser()
        parent = self.model(destinationType).load(
            destinationId, user=user, level=AccessType.ADMIN, exc=True)

        with ProgressContext(progress, user=user, title='Importing data') as ctx:
            return self.model('assetstore').importData(
                assetstore, parent=parent, parentType=destinationType, params={
                    'fileIncludeRegex': fileIncludeRegex,
                    'fileExcludeRegex': fileExcludeRegex,
                    'importPath': importPath,
                }, progress=ctx, user=user, leafFoldersAsItems=leafFoldersAsItems)

    @access.admin
    @autoDescribeRoute(
        Description('Update an existing assetstore.')
        .responseClass('Assetstore')
        .modelParam('id', model='assetstore')
        .param('name', 'Unique name for the assetstore.', strip=True)
        .param('root', 'Root path on disk (for Filesystem type)', required=False)
        .param('perms', 'File creation permissions (for Filesystem type)', required=False)
        .param('db', 'Database name (for GridFS type)', required=False)
        .param('mongohost', 'Mongo host URI (for GridFS type)', required=False)
        .param('replicaset', 'Replica set name (for GridFS type)', required=False)
        .param('bucket', 'The S3 bucket to store data in (for S3 type).', required=False)
        .param('prefix', 'Optional path prefix within the bucket under which '
               'files will be stored (for S3 type).', required=False, default='')
        .param('accessKeyId', 'The AWS access key ID to use for authentication '
               '(for S3 type).', required=False)
        .param('secret', 'The AWS secret key to use for authentication (for '
               'S3 type).', required=False)
        .param('service', 'The S3 service host (for S3 type).  Default is '
               's3.amazonaws.com.  This can be used to specify a protocol and '
               'port as well using the form [http[s]://](host domain)[:(port)]. '
               'Do not include the bucket name here.', required=False, default='')
        .param('readOnly', 'If this assetstore is read-only, set this to true.',
               required=False, dataType='boolean')
        .param('current', 'Whether this is the current assetstore', dataType='boolean')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def updateAssetstore(self, assetstore, name, root, perms, db, mongohost, replicaset, bucket,
                         prefix, accessKeyId, secret, service, readOnly, current, params):
        assetstore['name'] = name
        assetstore['current'] = current

        if assetstore['type'] == AssetstoreType.FILESYSTEM:
            self.requireParams({'root': root})
            assetstore['root'] = root
            if perms is not None:
                assetstore['perms'] = perms
        elif assetstore['type'] == AssetstoreType.GRIDFS:
            self.requireParams({'db': db})
            assetstore['db'] = db
            if mongohost is not None:
                assetstore['mongohost'] = mongohost
            if replicaset is not None:
                assetstore['replicaset'] = replicaset
        elif assetstore['type'] == AssetstoreType.S3:
            self.requireParams({
                'bucket': bucket,
                'accessKeyId': accessKeyId,
                'secret': secret
            })
            assetstore['bucket'] = bucket
            assetstore['prefix'] = prefix
            assetstore['accessKeyId'] = accessKeyId
            assetstore['secret'] = secret
            assetstore['service'] = service
            if readOnly is not None:
                assetstore['readOnly'] = readOnly
        else:
            event = events.trigger('assetstore.update', info={
                'assetstore': assetstore,
                'params': dict(
                    name=name, current=current, readOnly=readOnly, root=root, perms=perms,
                    db=db, mongohost=mongohost, replicaset=replicaset, bucket=bucket,
                    prefix=prefix, accessKeyId=accessKeyId, secret=secret, service=service,
                    **params
                )
            })
            if event.defaultPrevented:
                return
        return self.model('assetstore').save(assetstore)

    @access.admin
    @autoDescribeRoute(
        Description('Delete an assetstore.')
        .notes('This will fail if there are any files in the assetstore.')
        .modelParam('id', model='assetstore')
        .errorResponse(('A parameter was invalid.',
                        'The assetstore is not empty.'))
        .errorResponse('You are not an administrator.', 403)
    )
    def deleteAssetstore(self, assetstore, params):
        self.model('assetstore').remove(assetstore)
        return {'message': 'Deleted assetstore %s.' % assetstore['name']}

    @access.admin
    @autoDescribeRoute(
        Description('Get a list of files controlled by an assetstore.')
        .modelParam('id', model='assetstore')
        .pagingParams(defaultSort='_id')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403)
    )
    def getAssetstoreFiles(self, assetstore, limit, offset, sort, params):
        return list(self.model('file').find(
            query={'assetstoreId': assetstore['_id']},
            offset=offset, limit=limit, sort=sort))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

"""
The purpose of this module is to provide utility functions related to loading
and registering plugins into the system. The "loadPlugins" and "loadPlugin"
functions are used by the core system to actually load the plugins and create
the pseudo-packages required for easily referencing them. The other functions
in this class are general utility functions designed to be used within the
plugins themselves to help them register with the application.
"""

import codecs
import functools
import girder
import imp
import json
import os
import six
import sys
import yaml
import importlib

import pkg_resources
from pkg_resources import iter_entry_points

from girder import logprint
from girder.constants import GIRDER_ROUTE_ID, GIRDER_STATIC_ROUTE_ID, PACKAGE_DIR, ROOT_DIR, \
    ROOT_PLUGINS_PACKAGE, SettingKey
from girder.models.model_base import ValidationException
from girder.utility import mail_utils, model_importer

_pluginWebroots = {}


def loadPlugins(plugins, root, appconf, apiRoot=None, buildDag=True):
    """
    Loads a set of plugins into the application.

    :param plugins: The set of plugins to load, by directory name.
    :type plugins: list
    :param root: The root node of the server tree.
    :type root: object
    :param appconf: The server's cherrypy configuration object.
    :type appconf: dict
    :param apiRoot: The cherrypy api root object.
    :type apiRoot: object or None
    :param buildDag: If the ``plugins`` parameter is already a topo-sorted list
        with all dependencies resolved, set this to False and it will skip
        rebuilding the DAG. Otherwise the dependency resolution and sorting
        will occur within this method.
    :type buildDag: bool
    :returns: A 3-tuple containing the modified root, config, and apiRoot
        objects.
    :rtype tuple:
    """
    # Register a pseudo-package for the root of all plugins. This must be
    # present in the system module list in order to avoid import warnings.
    if ROOT_PLUGINS_PACKAGE not in sys.modules:
        module = imp.new_module(ROOT_PLUGINS_PACKAGE)
        girder.plugins = module
        sys.modules[ROOT_PLUGINS_PACKAGE] = module

    logprint.info('Resolving plugin dependencies...')

    if buildDag:
        plugins = getToposortedPlugins(plugins, ignoreMissing=True)

    for plugin in plugins:
        try:
            root, appconf, apiRoot = loadPlugin(plugin, root, appconf, apiRoot)
            logprint.success('Loaded plugin "%s"' % plugin)
        except Exception:
            logprint.exception('ERROR: Failed to load plugin "%s":' % plugin)

    return root, appconf, apiRoot


def getToposortedPlugins(plugins, ignoreMissing=False, keys=('dependencies',)):
    """
    Given a set of plugins to load, construct the full DAG of required plugins
    to load and yields them in toposorted order.

    :param ignoreMissing: Normally if one of the plugins specified does not exist,
        this raises a ValidationError. Set this to False to suppress that and instead
        print an error message and continue.
    :type ignoreMissing: bool
    :param keys: Keys that should be used to determine dependencies.
    :type keys: list of str
    """
    plugins = set(plugins)

    allPlugins = findAllPlugins()
    dag = {}
    visited = set()

    def addDeps(plugin):
        if plugin not in allPlugins:
            message = 'Required plugin %s does not exist.' % plugin
            if ignoreMissing:
                logprint.error(message)
                return
            else:
                raise ValidationException(message)

        deps = set()
        for key in keys:
            deps |= allPlugins[plugin][key]
        dag[plugin] = deps

        for dep in deps:
            if dep in visited:
                return
            visited.add(dep)
            addDeps(dep)

    for plugin in plugins:
        addDeps(plugin)

    for pset in toposort(dag):
        for plugin in pset:
            yield plugin


def loadPlugin(name, root, appconf, apiRoot=None):
    """
    Loads a plugin into the application. This means allowing it to create
    endpoints within its own web API namespace, and to register its event
    listeners, and do anything else it might want to do.

    :param name: The name of the plugin (i.e. its directory name)
    :type name: str
    :param root: The root node of the web API.
    :param appconf: The cherrypy configuration for the server.
    :type appconf: dict
    """
    if apiRoot is None:
        apiRoot = root.api.v1

    pluginDir = os.path.join(getPluginDir(), name)
    isPluginDir = os.path.isdir(os.path.join(pluginDir, 'server'))
    isPluginFile = os.path.isfile(os.path.join(pluginDir, 'server.py'))
    pluginLoadMethod = None

    if not os.path.exists(pluginDir):
        # Try to load the plugin as an entry_point
        for entry_point in iter_entry_points(group='girder.plugin', name=name):
            pluginLoadMethod = entry_point.load()
            module = importlib.import_module(entry_point.module_name)
            pluginDir = os.path.dirname(module.__file__)
            module.PLUGIN_ROOT_DIR = pluginDir
            girder.plugins.__dict__[name] = module
            isPluginDir = True

    if not os.path.exists(pluginDir):
        raise Exception('Plugin directory does not exist: %s' % pluginDir)
    if not isPluginDir and not isPluginFile:
        # This plugin does not have any server-side python code.
        return root, appconf, apiRoot

    mailTemplatesDir = os.path.join(pluginDir, 'server', 'mail_templates')
    if os.path.isdir(mailTemplatesDir):
        # If the plugin has mail templates, add them to the lookup path
        mail_utils.addTemplateDirectory(mailTemplatesDir, prepend=True)

    moduleName = '.'.join((ROOT_PLUGINS_PACKAGE, name))

    if moduleName not in sys.modules:
        fp = None
        try:
            # @todo this query is run for every plugin that's loaded
            setting = model_importer.ModelImporter().model('setting')
            routeTable = setting.get(SettingKey.ROUTE_TABLE)

            info = {
                'name': name,
                'config': appconf,
                'serverRoot': root,
                'serverRootPath': routeTable[GIRDER_ROUTE_ID],
                'apiRoot': apiRoot,
                'staticRoot': routeTable[GIRDER_STATIC_ROUTE_ID],
                'pluginRootDir': os.path.abspath(pluginDir)
            }

            if pluginLoadMethod is None:
                fp, pathname, description = imp.find_module(
                    'server', [pluginDir]
                )
                module = imp.load_module(moduleName, fp, pathname, description)
                module.PLUGIN_ROOT_DIR = pluginDir
                girder.plugins.__dict__[name] = module
                pluginLoadMethod = getattr(module, 'load', None)

            if pluginLoadMethod is not None:
                sys.modules[moduleName] = module
                pluginLoadMethod(info)

            root, appconf, apiRoot = (
                info['serverRoot'], info['config'], info['apiRoot'])

        finally:
            if fp:
                fp.close()

        return root, appconf, apiRoot


def getPluginDir():
    """
    Return the path to the directory that plugins are installed.
    """
    # Check if there is a plugins dir next to the girder dir.
    # This is the case when running from the git repository.
    pluginsDir = os.path.join(ROOT_DIR, 'plugins')
    if os.path.isdir(pluginsDir):
        return pluginsDir
    # Otherwise, we assume we are in a pip-installed environment where plugins is a subdir.
    else:
        return os.path.join(PACKAGE_DIR, 'plugins')


def findEntryPointPlugins(allPlugins):
    # look for plugins enabled via setuptools `entry_points`
    for entry_point in iter_entry_points(group='girder.plugin'):
        # set defaults
        allPlugins[entry_point.name] = {
            'name': entry_point.name,
            'description': '',
            'version': '',
            'dependencies': set()
        }
        configJson = os.path.join('girder', 'plugin.json')
        configYml = os.path.join('girder', 'plugin.yml')
        data = {}
        try:
            if pkg_resources.resource_exists(entry_point.name, configJson):
                with pkg_resources.resource_stream(
                        entry_point.name, configJson) as conf:
                    try:
                        data = json.load(codecs.getreader('utf8')(conf))
                    except ValueError:
                        logprint.exception(
                            'ERROR: Plugin "%s": plugin.json is not valid '
                            'JSON.' % entry_point.name)
            elif pkg_resources.resource_exists(entry_point.name, configYml):
                with pkg_resources.resource_stream(
                        entry_point.name, configYml) as conf:
                    try:
                        data = yaml.safe_load(conf)
                    except yaml.YAMLError:
                        logprint.exception(
                            'ERROR: Plugin "%s": plugin.yml is not valid '
                            'YAML.' % entry_point.name)
        except ImportError:
            pass
        if data == {}:
            data = getattr(entry_point.load(), 'config', {})
        allPlugins[entry_point.name].update(data)
        allPlugins[entry_point.name]['dependencies'] = set(
            allPlugins[entry_point.name]['dependencies'])


def findAllPlugins():
    """
    Walks the plugin directory to find all of the plugins. If the plugin has
    a plugin.json/yml file, this reads that file to determine dependencies.
    """
    allPlugins = {}

    findEntryPointPlugins(allPlugins)
    pluginDir = getPluginDir()

    for plugin in os.listdir(pluginDir):
        path = os.path.join(pluginDir, plugin)
        if not os.path.isdir(path):  # this also allows for symlinked directories
            continue

        data = {}
        configJson = os.path.join(path, 'plugin.json')
        configYml = os.path.join(path, 'plugin.yml')
        if os.path.isfile(configJson):
            with open(configJson) as conf:
                try:
                    data = json.load(conf)
                except ValueError:
                    logprint.exception(
                        'ERROR: Plugin "%s": plugin.json is not valid '
                        'JSON.' % plugin)
        elif os.path.isfile(configYml):
            with open(configYml) as conf:
                try:
                    data = yaml.safe_load(conf)
                except yaml.YAMLError:
                    logprint.exception(
                        'ERROR: Plugin "%s": plugin.yml is not valid '
                        'YAML.' % plugin)

        allPlugins[plugin] = {
            'name': data.get('name', plugin),
            'description': data.get('description', ''),
            'version': data.get('version', ''),
            'dependencies': set(data.get('dependencies', [])),
            'staticWebDependencies': set(data.get('staticWebDependencies', []))
        }
    return allPlugins


def toposort(data):
    """
    General-purpose topological sort function. Dependencies are expressed as a
    dictionary whose keys are items and whose values are a set of dependent
    items. Output is a list of sets in topological order. This is a generator
    function that returns a sequence of sets in topological order.

    :param data: The dependency information.
    :type data: dict
    :returns: Yields a list of sorted sets representing the sorted order.
    """
    if not data:
        return

    # Ignore self dependencies.
    for k, v in six.viewitems(data):
        v.discard(k)

    # Find all items that don't depend on anything.
    extra = functools.reduce(
        set.union, six.viewvalues(data)) - set(six.viewkeys(data))
    # Add empty dependencies where needed
    data.update({item: set() for item in extra})

    # Perform the topological sort.
    while True:
        ordered = set(item for item, dep in six.viewitems(data) if not dep)
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered)
                for item, dep in six.viewitems(data) if item not in ordered}
    # Detect any cycles in the dependency graph.
    if data:
        raise Exception('Cyclic dependencies detected:\n%s' % '\n'.join(
                        repr(x) for x in six.viewitems(data)))


def addChildNode(node, name, obj=None):
    """
    Use this to build paths to your plugin's endpoints.

    :param node: The parent node to add the child node to.
    :param name: The name of the child node in the URL path.
    :type name: str
    :param obj: The object to place at this new node, or None if this child
                should not be exposed as an endpoint, instead just used as
                an intermediary hidden node.
    :type obj: object or None
    :returns: The node that was created.
    """
    if obj:
        setattr(node, name, obj)
        return obj
    else:
        hiddenNode = type('', (), dict(exposed=False))()
        setattr(node, name, hiddenNode)
        return hiddenNode


def getPluginWebroots():
    global _pluginWebroots
    return _pluginWebroots


def registerPluginWebroot(webroot, name):
    """
    Adds a webroot to the global registry for plugins based on
    the plugin name.
    """
    global _pluginWebroots

    _pluginWebroots[name] = webroot


class config(object):  # noqa: class name
    """
    Wrap a plugin's ``load`` method appending plugin metadata.

    :param name str: The plugin's name
    :param description str: A brief description of the plugin.
    :param version str: A semver compatible version string.
    :param dependencies list: A list of plugins required by this plugin.
    :param python3 bool: Whether this plugin supports python3.
    :returns: A decorator that appends the metadata to the method
    """
    def __init__(self, **kw):
        self.config = kw

    def __call__(self, func):
        @six.wraps(func)
        def wrapped(*arg, **kw):
            return func(*arg, **kw)
        wrapped.config = self.config
        return wrapped



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from ..describe import Description, autoDescribeRoute
from ..rest import Resource, RestException, filtermodel, setResponseHeader
from girder.utility import ziputil
from girder.constants import AccessType, TokenScope
from girder.api import access


class Item(Resource):

    def __init__(self):
        super(Item, self).__init__()
        self.resourceName = 'item'
        self.route('DELETE', (':id',), self.deleteItem)
        self.route('GET', (), self.find)
        self.route('GET', (':id',), self.getItem)
        self.route('GET', (':id', 'files'), self.getFiles)
        self.route('GET', (':id', 'download'), self.download)
        self.route('GET', (':id', 'rootpath'), self.rootpath)
        self.route('POST', (), self.createItem)
        self.route('PUT', (':id',), self.updateItem)
        self.route('POST', (':id', 'copy'), self.copyItem)
        self.route('PUT', (':id', 'metadata'), self.setMetadata)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('List or search for items.')
        .responseClass('Item', array=True)
        .param('folderId', 'Pass this to list all items in a folder.',
               required=False)
        .param('text', 'Pass this to perform a full text search for items.',
               required=False)
        .param('name', 'Pass to lookup an item by exact name match. Must '
               'pass folderId as well when using this.', required=False)
        .pagingParams(defaultSort='lowerName')
        .errorResponse()
        .errorResponse('Read access was denied on the parent folder.', 403)
    )
    def find(self, folderId, text, name, limit, offset, sort, params):
        """
        Get a list of items with given search parameters. Currently accepted
        search modes are:

        1. Searching by folderId, with optional additional filtering by the name
           field (exact match) or using full text search within a single parent
           folder. Pass a "name" parameter or "text" parameter to invoke these
           additional filters.
        2. Searching with full text search across all items in the system.
           Simply pass a "text" parameter for this mode.
        """
        user = self.getCurrentUser()

        if folderId:
            folder = self.model('folder').load(
                id=folderId, user=user, level=AccessType.READ, exc=True)
            filters = {}
            if text:
                filters['$text'] = {
                    '$search': text
                }
            if name:
                filters['name'] = name

            return list(self.model('folder').childItems(
                folder=folder, limit=limit, offset=offset, sort=sort, filters=filters))
        elif text is not None:
            return list(self.model('item').textSearch(
                text, user=user, limit=limit, offset=offset, sort=sort))
        else:
            raise RestException('Invalid search mode.')

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('Get an item by ID.')
        .responseClass('Item')
        .modelParam('id', model='item', level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the item.', 403)
    )
    def getItem(self, item, params):
        return item

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('Create a new item.')
        .responseClass('Item')
        .modelParam('folderId', 'The ID of the parent folder.',
                    level=AccessType.WRITE, paramType='query')
        .param('name', 'Name for the item.', strip=True)
        .param('description', 'Description for the item.', required=False,
               default='', strip=True)
        .param('reuseExisting', 'Return existing item (by name) if it exists.',
               required=False, dataType='boolean', default=False)
        .errorResponse()
        .errorResponse('Write access was denied on the parent folder.', 403)
    )
    def createItem(self, folder, name, description, reuseExisting, params):
        return self.model('item').createItem(
            folder=folder, name=name, creator=self.getCurrentUser(), description=description,
            reuseExisting=reuseExisting)

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('Edit an item or move it to another folder.')
        .responseClass('Item')
        .modelParam('id', model='item', level=AccessType.WRITE)
        .param('name', 'Name for the item.', required=False, strip=True)
        .param('description', 'Description for the item.', required=False)
        .modelParam('folderId', 'Pass this to move the item to a new folder.',
                    required=False, paramType='query', level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the item or folder.', 403)
    )
    def updateItem(self, item, name, description, folder, params):
        if name is not None:
            item['name'] = name
        if description is not None:
            item['description'] = description

        self.model('item').updateItem(item)

        if folder and folder['_id'] != item['folderId']:
            self.model('item').move(item, folder)

        return item

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('Set metadata fields on an item.')
        .responseClass('Item')
        .notes('Set metadata fields to null in order to delete them.')
        .modelParam('id', model='item', level=AccessType.WRITE)
        .jsonParam('metadata', 'A JSON object containing the metadata keys to add',
                   paramType='body')
        .errorResponse(('ID was invalid.',
                        'Invalid JSON passed in request body.',
                        'Metadata key name was invalid.'))
        .errorResponse('Write access was denied for the item.', 403)
    )
    def setMetadata(self, item, metadata, params):
        # Make sure we let user know if we can't accept a metadata key
        for k in metadata:
            if not k:
                raise RestException('Key names must not be empty.')
            if '.' in k:
                raise RestException(
                    'Invalid key %s: keys must not contain the "." character.' % k)
            if k[0] == '$':
                raise RestException(
                    'Invalid key %s: keys must not start with the "$" character.' % k)

        return self.model('item').setMetadata(item, metadata)

    def _downloadMultifileItem(self, item, user):
        setResponseHeader('Content-Type', 'application/zip')
        setResponseHeader(
            'Content-Disposition',
            'attachment; filename="%s%s"' % (item['name'], '.zip'))

        def stream():
            zip = ziputil.ZipGenerator(item['name'])
            for (path, file) in self.model('item').fileList(item, subpath=False):
                for data in zip.addFile(file, path):
                    yield data
            yield zip.footer()
        return stream

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model='file')
    @autoDescribeRoute(
        Description('Get the files within an item.')
        .responseClass('File', array=True)
        .modelParam('id', model='item', level=AccessType.READ)
        .pagingParams(defaultSort='name')
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the item.', 403)
    )
    def getFiles(self, item, limit, offset, sort, params):
        return list(self.model('item').childFiles(
            item=item, limit=limit, offset=offset, sort=sort))

    @access.cookie
    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Download the contents of an item.')
        .modelParam('id', model='item', level=AccessType.READ)
        .param('offset', 'Byte offset into the file.', dataType='int', default=0)
        .param('format', 'If unspecified, items with one file are downloaded '
               'as that file, and other items are downloaded as a zip '
               'archive.  If \'zip\', a zip archive is always sent.',
               required=False)
        .param('contentDisposition', 'Specify the Content-Disposition response '
               'header disposition-type value, only applied for single file '
               'items.', required=False, enum=['inline', 'attachment'],
               default='attachment')
        .param('extraParameters', 'Arbitrary data to send along with the '
               'download request, only applied for single file '
               'items.', required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the item.', 403)
    )
    def download(self, item, offset, format, contentDisposition, extraParameters, params):
        user = self.getCurrentUser()
        files = list(self.model('item').childFiles(item=item, limit=2))
        if format not in (None, '', 'zip'):
            raise RestException('Unsupported format: %s.' % format)
        if len(files) == 1 and format != 'zip':
            if contentDisposition not in {None, 'inline', 'attachment'}:
                raise RestException(
                    'Unallowed contentDisposition type "%s".' % contentDisposition)
            return self.model('file').download(
                files[0], offset, contentDisposition=contentDisposition,
                extraParameters=extraParameters)
        else:
            return self._downloadMultifileItem(item, user)

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Delete an item by ID.')
        .modelParam('id', model='item', level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Write access was denied for the item.', 403)
    )
    def deleteItem(self, item, params):
        self.model('item').remove(item)
        return {'message': 'Deleted item %s.' % item['name']}

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get the path to the root of the item\'s hierarchy.')
        .modelParam('id', model='item', level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the item.', 403)
    )
    def rootpath(self, item, params):
        return self.model('item').parentsToRoot(item, self.getCurrentUser())

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model='item')
    @autoDescribeRoute(
        Description('Copy an item.')
        .notes('If no folderId parameter is specified, creates a copy of the item in '
               'its current containing folder.')
        .responseClass('Item')
        .modelParam('id', 'The ID of the original item.', model='item', level=AccessType.READ)
        .modelParam('folderId', 'The ID of the parent folder.', required=False,
                    level=AccessType.WRITE)
        .param('name', 'Name for the new item.', required=False, strip=True)
        .param('description', 'Description for the new item.', required=False, strip=True)
        .errorResponse(('A parameter was invalid.',
                        'ID was invalid.'))
        .errorResponse('Read access was denied on the original item.\n\n'
                       'Write access was denied on the parent folder.', 403)
    )
    def copyItem(self, item, folder, name, description, params):
        user = self.getCurrentUser()

        if folder is None:
            folder = self.model('folder').load(
                id=item['folderId'], user=user, level=AccessType.WRITE, exc=True)
        return self.model('item').copyItem(
            item, creator=user, name=name, folder=folder, description=description)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2016 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from bson.objectid import ObjectId
from girder import events
from girder.api import access
from girder.api.rest import getCurrentToken, setCurrentUser
from girder.utility import mail_utils
from girder.utility.model_importer import ModelImporter

from .constants import TOKEN_SCOPE_AUTHORIZED_UPLOAD
from .rest import AuthorizedUpload


@access.public
def _authorizeInitUpload(event):
    """
    Called when initializing an upload, prior to the default handler. Checks if
    the user is passing an authorized upload token, and if so, sets the current
    request-thread user to be whoever created the token.
    """
    token = getCurrentToken()
    params = event.info['params']
    tokenModel = ModelImporter.model('token')
    parentType = params.get('parentType')
    parentId = params.get('parentId', '')
    requiredScopes = {TOKEN_SCOPE_AUTHORIZED_UPLOAD, 'authorized_upload_folder_%s' % parentId}

    if parentType == 'folder' and tokenModel.hasScope(token=token, scope=requiredScopes):
        user = ModelImporter.model('user').load(token['userId'], force=True)
        setCurrentUser(user)


def _storeUploadId(event):
    """
    Called after an upload is first initialized successfully. Sets the authorized upload ID
    in the token, ensuring it can be used for only this upload.
    """
    returnVal = event.info['returnVal']
    token = getCurrentToken()
    tokenModel = ModelImporter.model('token')
    isAuthorizedUpload = tokenModel.hasScope(token, TOKEN_SCOPE_AUTHORIZED_UPLOAD)

    if isAuthorizedUpload and returnVal.get('_modelType', 'upload') == 'upload':
        params = event.info['params']
        token['scope'].remove(TOKEN_SCOPE_AUTHORIZED_UPLOAD)
        token['authorizedUploadId'] = returnVal['_id']
        token['authorizedUploadDescription'] = params.get('authorizedUploadDescription', '')
        token['authorizedUploadEmail'] = params.get('authorizedUploadEmail')
        tokenModel.save(token)


@access.public
def _authorizeUploadStep(event):
    """
    Called before any requests dealing with partially completed uploads. Sets the
    request thread user to the authorized upload token creator if the requested
    upload is an authorized upload.
    """
    token = getCurrentToken()
    uploadId = ObjectId(event.info['params'].get('uploadId'))

    if token and 'authorizedUploadId' in token and token['authorizedUploadId'] == uploadId:
        user = ModelImporter.model('user').load(token['userId'], force=True)
        setCurrentUser(user)


def _uploadComplete(event):
    """
    Called after an upload finishes. We check if our current token is a special
    authorized upload token, and if so, delete it.

    TODO we could alternatively keep a reference count inside each token that authorized
    more than a single upload at a time, and just decrement it here.
    """
    token = getCurrentToken()
    if 'authorizedUploadId' in token:
        user = ModelImporter.model('user').load(token['userId'], force=True)
        item = ModelImporter.model('item').load(event.info['file']['itemId'], force=True)

        # Save the metadata on the item
        item['description'] = token['authorizedUploadDescription']
        item['authorizedUploadEmail'] = token['authorizedUploadEmail']
        ModelImporter.model('item').save(item)

        text = mail_utils.renderTemplate('authorized_upload.uploadFinished.mako', {
            'itemId': item['_id'],
            'itemName': item['name'],
            'itemDescription': item.get('description', '')
        })
        mail_utils.sendEmail(to=user['email'], subject='Authorized upload complete', text=text)
        ModelImporter.model('token').remove(token)


def load(info):
    name = info['name']

    events.bind('rest.post.file.before', name, _authorizeInitUpload)
    events.bind('rest.post.file.after', name, _storeUploadId)
    events.bind('rest.post.file/chunk.before', name, _authorizeUploadStep)
    events.bind('rest.post.file/completion.before', name, _authorizeUploadStep)
    events.bind('rest.get.file/offset.before', name, _authorizeUploadStep)
    events.bind('model.file.finalizeUpload.after', name, _uploadComplete)

    info['apiRoot'].authorized_upload = AuthorizedUpload()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder import events
from girder.constants import AccessType
from girder.utility.model_importer import ModelImporter
from . import rest


def removeThumbnails(event):
    """
    When a resource containing thumbnails is about to be deleted, we delete
    all of the thumbnails that are attached to it.
    """
    thumbs = event.info.get('_thumbnails', ())
    fileModel = ModelImporter.model('file')

    for fileId in thumbs:
        file = fileModel.load(fileId, force=True)
        if file:
            fileModel.remove(file)


def removeThumbnailLink(event):
    """
    When a thumbnail file is deleted, we remove the reference to it from the
    resource to which it is attached.
    """
    doc = event.info

    if doc.get('isThumbnail'):
        model = ModelImporter.model(doc['attachedToType'])
        resource = model.load(doc['attachedToId'], force=True)

        if doc['_id'] in resource.get('_thumbnails', ()):
            resource['_thumbnails'].remove(doc['_id'])
            model.save(resource, validate=False)


def load(info):
    info['apiRoot'].thumbnail = rest.Thumbnail()

    for model in ('item', 'collection', 'folder', 'user'):
        ModelImporter.model(model).exposeFields(
            level=AccessType.READ, fields='_thumbnails')

        events.bind('model.%s.remove' % model, 'thumbnails', removeThumbnails)

    events.bind('model.file.remove', 'thumbnails', removeThumbnailLink)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from girder import events
from girder.constants import AssetstoreType

from . metadata_extractor import ServerMetadataExtractor


def handler(event):
    if event.info['assetstore']['type'] == AssetstoreType.FILESYSTEM:
        metadataExtractor = ServerMetadataExtractor(event.info['assetstore'],
                                                    event.info['file'])
        metadataExtractor.extractMetadata()


def load(info):
    events.bind('data.process', 'metadata_extractor_handler', handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2014 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import copy
import datetime
import six

from bson.objectid import ObjectId
from girder import events
from girder.api import access
from girder.api.describe import Description, describeRoute
from girder.api.rest import Resource, RestException
from girder.constants import AccessType
from girder.models.model_base import AccessControlledModel
from girder.utility import acl_mixin
from . import constants


class ResourceExt(Resource):
    def __init__(self, info):
        """
        Initialize the resource.  This saves the info path so that we can
        dynamically change the routes when the setting listing the resources we
        use changes.
        :param info: the info class passed to the load function.
        """
        super(ResourceExt, self).__init__()
        self.loadInfo = info
        self.boundResources = {}

    # These methods implement the endpoint routing

    def bindModels(self, event=None):
        """
        When the list of tracked provenance resources is changed, rebind the
        appropriate models to this instance of this class.
        :param event: the event when a setting is saved, or None to check the
                      binding.
        """
        if not event or event.name == 'provenance.initialize':
            pass
        elif event.name == 'model.setting.save.after':
            if not hasattr(event, "info"):
                return
            if (event.info.get('key', '') !=
                    constants.PluginSettings.PROVENANCE_RESOURCES):
                return
        else:
            return
        resources = self.model('setting').get(
            constants.PluginSettings.PROVENANCE_RESOURCES)
        if resources:
            resources = resources.replace(',', ' ').strip().split()
        else:
            resources = []
        resources = dict.fromkeys(resources)
        # Always include item
        resources['item'] = None
        # Exclude resources that should never have provenance
        for disallowedResource in ('model_base', 'notification', 'password',
                                   'token'):
            if disallowedResource in resources:
                del resources[disallowedResource]
        self.unbindModels(resources)
        for resource in resources:
            if resource not in self.boundResources:
                events.bind('model.%s.save' % resource, 'provenance',
                            self.resourceSaveHandler)
                events.bind('model.%s.copy.prepare' % resource,
                            'provenance', self.resourceCopyHandler)
                if hasattr(self.loadInfo['apiRoot'], resource):
                    getattr(self.loadInfo['apiRoot'], resource).route(
                        'GET', (':id', 'provenance'),
                        self.getGetHandler(resource))
                self.boundResources[resource] = True

    def unbindModels(self, resources={}):
        """
        Unbind any models that were bound and aren't listed as needed.
        :param resources: resources that shouldn't be unbound.
        """
        # iterate over a list so that we can change the dictionary as we use it
        for oldresource in list(six.viewkeys(self.boundResources)):
            if oldresource not in resources:
                # Unbind this and remove it from the api
                events.unbind('model.%s.save' % oldresource, 'provenance')
                events.unbind('model.%s.copy.prepare' % oldresource,
                              'provenance')
                if hasattr(self.loadInfo['apiRoot'], oldresource):
                    getattr(self.loadInfo['apiRoot'], oldresource).removeRoute(
                        'GET', (':id', 'provenance'),
                        self.getGetHandler(oldresource))
                del self.boundResources[oldresource]

    def getGetHandler(self, resource):
        """
        Return a function that will get the provenance for a particular
        resource type.  This creates such a function if necessary, copying the
        main function and setting an internal value so that the function is
        coupled to the resource.
        :param resource: the name of the resource to get.
        :returns: getHandler function.
        """
        key = 'provenanceGetHandler_%s' % resource
        if not hasattr(self, key):
            def resourceGetHandler(id, params):
                return self.provenanceGetHandler(id, params, resource)
            # We inherit the description and access decorator details from the
            # general provenanceGetHandler
            for attr in ('description', 'accessLevel'):
                setattr(resourceGetHandler, attr,
                        getattr(self.provenanceGetHandler, attr))
            setattr(self, key, resourceGetHandler)
        return getattr(self, key)

    @access.public
    @describeRoute(
        Description('Get the provenance for a given resource.')
        .param('id', 'The resource ID', paramType='path')
        .param('version', 'The provenance version for the resource.  If not '
               'specified, the latest provenance data is returned.  If "all" '
               'is specified, a list of all provenance data is returned.  '
               'Negative indices can also be used (-1 is the latest '
               'provenance, -2 second latest, etc.).', required=False)
        .errorResponse()
    )
    def provenanceGetHandler(self, id, params, resource=None):
        user = self.getCurrentUser()
        model = self.model(resource)
        if isinstance(model, (acl_mixin.AccessControlMixin,
                              AccessControlledModel)):
            obj = model.load(id, level=AccessType.READ, user=user)
        else:
            obj = model.load(id)
        version = -1
        if 'version' in params:
            if params['version'] == 'all':
                version = None
            else:
                try:
                    version = int(params['version'])
                except ValueError:
                    raise RestException('Invalid version.')
        provenance = obj.get('provenance', [])
        result = None
        if version is None or version == 0:
            result = provenance
        elif version < 0:
            if len(provenance) >= -version:
                result = provenance[version]
        else:
            for prov in provenance:
                if prov.get('version', None) == version:
                    result = prov
                    break
        return {
            'resourceId': id,
            'provenance': result
        }

    # These methods maintain the provenance

    def resourceSaveHandler(self, event):
        # get the resource name from the event
        resource = event.name.split('.')[1]
        obj = event.info
        if ('_id' not in obj or ('provenance' not in obj and
                                 obj.get('updated', None) ==
                                 obj.get('created', 'unknown'))):
            self.createNewProvenance(obj, resource)
        else:
            if 'provenance' not in obj:
                self.createExistingProvenance(obj, resource)
            elif (obj.get('updated', None) !=
                    obj['provenance'][-1].get('eventTime', False)):
                self.updateProvenance(obj, resource)

    def createNewProvenance(self, obj, resource):
        provenance = []
        created = obj.get('created', datetime.datetime.utcnow())
        creatorId = obj.get('creatorId', None)
        if creatorId is None:
            user = self.getProvenanceUser(obj)
            if user is not None:
                creatorId = user['_id']
        creationEvent = {
            'eventType': 'creation',
            'eventUser': creatorId,
            'eventTime': obj.get('updated', created),
            'created': created
        }
        obj['provenance'] = provenance
        self.addProvenanceEvent(obj, creationEvent, resource)

    def getProvenanceUser(self, obj):
        """
        Get the user that is associated with the current provenance change.
        This is the current session user, if there is one.  If not, it is the
        object's user or creator.
        :param obj: a model object.
        :returns: user for the object or None.
        """
        user = self.getCurrentUser()
        if obj and not user:
            user = obj.get('userId', None)
            if not user:
                user = obj.get('creatorId', None)
        if isinstance(user, tuple([ObjectId] + list(six.string_types))):
            user = self.model('user').load(user, force=True)
        return user

    def createExistingProvenance(self, obj, resource):
        self.createNewProvenance(obj, resource)
        # we don't know what happened between creation and now
        provenance = obj['provenance']
        provenance[0]['eventType'] = 'unknownHistory'
        # but we can track starting now
        self.updateProvenance(obj, resource)

    def addProvenanceEvent(self, obj, provenanceEvent, resource):
        if 'provenance' not in obj:
            self.createExistingProvenance(obj, resource)
        provenance = obj['provenance']
        self.incrementVersion(provenance, provenanceEvent)
        provenance.append(provenanceEvent)

    def incrementVersion(self, provenance, provenanceEvent):
        if len(provenance) == 0:
            # counting from 1, since people do that
            provenanceEvent['version'] = 1
        else:
            provenanceEvent['version'] = int(provenance[-1]['version']) + 1

    def updateProvenance(self, curObj, resource):
        """
        Update the provenance record of an object.
        :param curObj: the object to potentially update.
        :param resource: the type of resource (model name).
        :returns: True if the provenance was updated, False if it stayed the
                  same.
        """
        user = self.getProvenanceUser(curObj)
        model = self.model(resource)
        if isinstance(model, (acl_mixin.AccessControlMixin,
                              AccessControlledModel)):
            prevObj = model.load(curObj['_id'], force=True)
        else:
            prevObj = model.load(curObj['_id'])
        if prevObj is None:
            return False
        oldData, newData = self.resourceDifference(prevObj, curObj)
        if not len(newData) and not len(oldData):
            return False
        updateEvent = {
            'eventType': 'update',
            'eventTime': curObj.get('updated', datetime.datetime.utcnow()),
            'new': newData,
            'old': oldData
        }
        if user is not None:
            updateEvent['eventUser'] = user['_id']
        self.addProvenanceEvent(curObj, updateEvent, resource)
        return True

    def resourceDifference(self, prevObj, curObj):
        """
        Generate dictionaries with values that have changed between two
        objects.
        :param prevObj: the initial state of the object.
        :param curObj: the current state of the object.
        :returns: oldData: a dictionary of values that are no longer the same
                           or present in the object.
        :returns: newData: a dictionary of values that are different or new in
                           the object.
        """
        curSnapshot = self.snapshotResource(curObj)
        prevSnapshot = self.snapshotResource(prevObj)
        oldData = {}
        newData = {}
        for key in curSnapshot:
            if key in prevSnapshot:
                try:
                    if curSnapshot[key] != prevSnapshot[key]:
                        newData[key] = curSnapshot[key]
                        oldData[key] = prevSnapshot[key]
                except TypeError:
                    # If the data types of the old and new keys are not
                    # comparable, an error is thrown.  In this case, always
                    # treat them as different.
                    newData[key] = curSnapshot[key]
                    oldData[key] = prevSnapshot[key]
            else:
                newData[key] = curSnapshot[key]
        for key in prevSnapshot:
            if key not in curSnapshot:
                oldData[key] = prevSnapshot[key]
        return oldData, newData

    def snapshotResource(self, obj):
        """
        Generate a dictionary that represents an arbitrary resource.  All
        fields are included except provenance and values starting with _.
        :param obj: the object for which to generate a snapshop dictionary.
        :param includeItemFiles: if True and this is an item, include files.
        :returns: a snapshot dictionary.
        """
        ignoredKeys = ('provenance', 'updated')
        snap = {key: obj[key] for key in obj
                if not key.startswith('_') and key not in ignoredKeys}
        return snap

    def fileSaveHandler(self, event):
        """
        When a file is saved, update the provenance of the parent item.
        :param event: the event with the file information.
        """
        curFile = event.info
        if not curFile.get('itemId') or '_id' not in curFile:
            return
        user = self.getProvenanceUser(curFile)
        item = self.model('item').load(id=curFile['itemId'], force=True)
        if not item:
            return
        prevFile = self.model('file').load(curFile['_id'], force=True)
        if prevFile is None:
            oldData = None
            newData = self.snapshotResource(curFile)
        else:
            oldData, newData = self.resourceDifference(prevFile, curFile)
        if not len(newData) and not len(oldData):
            return
        updateEvent = {
            'eventType': 'fileUpdate',
            'eventTime': curFile.get('updated', datetime.datetime.utcnow()),
            'file': [{
                'fileId': curFile['_id'],
                'new': newData
            }]
        }
        if oldData is not None:
            updateEvent['file'][0]['old'] = oldData
        if user is not None:
            updateEvent['eventUser'] = user['_id']
        self.addProvenanceEvent(item, updateEvent, 'item')
        self.model('item').save(item, triggerEvents=False)

    def fileSaveCreatedHandler(self, event):
        """
        When a file is created, we don't record it in the save handler
        because we want to know its id.  We record it here, instead.
        :param event: the event with the file information.
        """
        file = event.info
        if not file.get('itemId') or '_id' not in file:
            return
        user = self.getProvenanceUser(file)
        item = self.model('item').load(id=file['itemId'], force=True)
        if not item:
            return
        updateEvent = {
            'eventType': 'fileAdded',
            'eventTime': file.get('created', datetime.datetime.utcnow()),
            'file': [{
                'fileId': file['_id'],
                'new': self.snapshotResource(file)
            }]
        }
        if user is not None:
            updateEvent['eventUser'] = user['_id']
        self.addProvenanceEvent(item, updateEvent, 'item')
        self.model('item').save(item, triggerEvents=False)

    def fileRemoveHandler(self, event):
        """
        When a file is removed, update the provenance of the parent item.
        :param event: the event with the file information.
        """
        file = event.info
        itemId = file.get('itemId')
        # Don't attach provenance to an item based on files that are not
        # directly associated (we may want to revisit this and, when files are
        # attachedToType, add provenance to the appropriate type and ID).
        if not itemId:
            return
        user = self.getProvenanceUser(file)
        item = self.model('item').load(id=itemId, force=True)
        if not item:
            return
        updateEvent = {
            'eventType': 'fileRemoved',
            'eventTime': datetime.datetime.utcnow(),
            'file': [{
                'fileId': file['_id'],
                'old': self.snapshotResource(file)
            }]
        }
        if user is not None:
            updateEvent['eventUser'] = user['_id']
        self.addProvenanceEvent(item, updateEvent, 'item')
        self.model('item').save(item, triggerEvents=False)

    def resourceCopyHandler(self, event):
        # Use the old item's provenance, but add a copy record.
        resource = event.name.split('.')[1]
        srcObj, newObj = event.info
        # We should have marked the new object already when it was first
        # created.  If not, exit
        if 'provenance' not in newObj:
            pass  # pragma: no cover
        if 'provenance' in srcObj:
            newProv = newObj['provenance'][-1]
            newObj['provenance'] = copy.deepcopy(srcObj['provenance'])
            newProv['version'] = newObj['provenance'][-1]['version'] + 1
            newObj['provenance'].append(newProv)
        # Convert the creation record to a copied record
        newObj['provenance'][-1]['eventType'] = 'copy'
        if '_id' in srcObj:
            newObj['provenance'][-1]['originalId'] = srcObj['_id']
        self.model(resource).save(newObj, triggerEvents=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import six

from girder.api import rest
from girder.models.model_base import AccessException


def admin(*args, **kwargs):
    """
    Functions that require administrator access should be wrapped in this
    decorator.

    :param scope: To also expose this endpoint for certain token scopes,
        pass those scopes here. If multiple are passed, all will be required.
    :type scope: str or list of str
    """
    if len(args) == 1 and callable(args[0]):  # Raw decorator
        @six.wraps(args[0])
        def wrapped(*iargs, **ikwargs):
            rest.requireAdmin(rest.getCurrentUser())
            return args[0](*iargs, **ikwargs)
        wrapped.accessLevel = 'admin'
        return wrapped
    else:  # We should return a decorator
        def dec(fun):
            @six.wraps(fun)
            def wrapped(*iargs, **ikwargs):
                rest.requireAdmin(rest.getCurrentUser())
                return fun(*iargs, **ikwargs)
            wrapped.accessLevel = 'admin'
            wrapped.requiredScopes = kwargs.get('scope')
            return wrapped
        return dec


def user(*args, **kwargs):
    """
    Functions that require a logged-in user should be wrapped with this access
    decorator.

    :param scope: To also expose this endpoint for certain token scopes,
        pass those scopes here. If multiple are passed, all will be required.
    :type scope: str or list of str
    """
    if len(args) == 1 and callable(args[0]):  # Raw decorator
        @six.wraps(args[0])
        def wrapped(*iargs, **ikwargs):
            if not rest.getCurrentUser():
                raise AccessException('You must be logged in.')
            return args[0](*iargs, **ikwargs)
        wrapped.accessLevel = 'user'
        return wrapped
    else:  # We should return a decorator
        def dec(fun):
            @six.wraps(fun)
            def wrapped(*iargs, **ikwargs):
                if not rest.getCurrentUser():
                    raise AccessException('You must be logged in.')
                return fun(*iargs, **ikwargs)
            wrapped.accessLevel = 'user'
            wrapped.requiredScopes = kwargs.get('scope')
            return wrapped
        return dec


def token(*args, **kwargs):
    """
    Functions that require a token, but not necessarily a user authentication
    token, should use this access decorator.

    :param scope: The scope or list of scopes required for this token.
    :type scope: str or list of str
    """
    if len(args) == 1 and callable(args[0]):  # Raw decorator
        @six.wraps(args[0])
        def wrapped(*iargs, **ikwargs):
            if not rest.getCurrentToken():
                raise AccessException(
                    'You must be logged in or have a valid auth token.')
            return args[0](*iargs, **ikwargs)
        wrapped.accessLevel = 'token'
        return wrapped
    else:  # We should return a decorator
        def dec(fun):
            @six.wraps(fun)
            def wrapped(*iargs, **ikwargs):
                if not rest.getCurrentToken():
                    raise AccessException(
                        'You must be logged in or have a valid auth token.')
                return fun(*iargs, **ikwargs)
            wrapped.accessLevel = 'token'
            wrapped.requiredScopes = kwargs.get('scope')
            return wrapped
        return dec


def public(*args, **kwargs):
    """
    Functions that allow any client access, including those that haven't logged
    in should be wrapped in this decorator.
    """
    if len(args) == 1 and callable(args[0]):  # Raw decorator
        args[0].accessLevel = 'public'
        return args[0]
    else:  # We should return a decorator
        def dec(fun):
            fun.accessLevel = 'public'
            fun.requiredScopes = kwargs.get('scope')
            return fun
        return dec


def cookie(*args, **kwargs):
    """
    REST endpoints that allow the use of a cookie for authentication should be
    wrapped in this decorator.

    When used as a normal decorator, this is only effective on endpoints for
    HEAD and GET routes (as these routes should be read-only in a RESTful API).

    While allowing cookie authentication on other types of routes exposes an
    application to Cross-Site Request Forgery (CSRF) attacks, an optional
    ``force=True`` kwarg may be passed to the decorator to make it effective
    on any type of route.
    """
    if len(args) == 1 and callable(args[0]):  # Used as a raw decorator
        force = False
        args[0].cookieAuth = (True, force)
        return args[0]
    else:  # Used with arguments
        def decorator(fun):
            fun.cookieAuth = (True, kwargs.get('force', False))
            return fun

        return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

from ..describe import Description, autoDescribeRoute
from ..rest import Resource, filtermodel, setResponseHeader
from girder.api import access
from girder.constants import AccessType, TokenScope
from girder.models.model_base import AccessException
from girder.utility import ziputil
from girder.utility.progress import ProgressContext


class Collection(Resource):
    """API Endpoint for collections."""
    def __init__(self):
        super(Collection, self).__init__()
        self.resourceName = 'collection'
        self.route('DELETE', (':id',), self.deleteCollection)
        self.route('GET', (), self.find)
        self.route('GET', (':id',), self.getCollection)
        self.route('GET', (':id', 'details'), self.getCollectionDetails)
        self.route('GET', (':id', 'download'), self.downloadCollection)
        self.route('GET', (':id', 'access'), self.getCollectionAccess)
        self.route('POST', (), self.createCollection)
        self.route('PUT', (':id',), self.updateCollection)
        self.route('PUT', (':id', 'access'), self.updateCollectionAccess)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model='collection')
    @autoDescribeRoute(
        Description('List or search for collections.')
        .responseClass('Collection', array=True)
        .param('text', 'Pass this to perform a text search for collections.', required=False)
        .pagingParams(defaultSort='name')
    )
    def find(self, text, limit, offset, sort, params):
        user = self.getCurrentUser()

        if text is not None:
            return list(self.model('collection').textSearch(
                text, user=user, limit=limit, offset=offset))

        return list(self.model('collection').list(
            user=user, offset=offset, limit=limit, sort=sort))

    @access.user(scope=TokenScope.DATA_WRITE)
    @filtermodel(model='collection')
    @autoDescribeRoute(
        Description('Create a new collection.')
        .responseClass('Collection')
        .param('name', 'Name for the collection. Must be unique.')
        .param('description', 'Collection description.', required=False)
        .param('public', 'Whether the collection should be publicly visible.',
               required=False, dataType='boolean', default=False)
        .errorResponse()
        .errorResponse('You are not authorized to create collections.', 403)
    )
    def createCollection(self, name, description, public, params):
        user = self.getCurrentUser()

        if not self.model('collection').hasCreatePrivilege(user):
            raise AccessException('You are not authorized to create collections.')

        return self.model('collection').createCollection(
            name=name, description=description, public=public, creator=user)

    @access.public(scope=TokenScope.DATA_READ)
    @filtermodel(model='collection')
    @autoDescribeRoute(
        Description('Get a collection by ID.')
        .responseClass('Collection')
        .modelParam('id', model='collection', level=AccessType.READ)
        .errorResponse('ID was invalid.')
        .errorResponse('Read permission denied on the collection.', 403)
    )
    def getCollection(self, collection, params):
        return collection

    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Get detailed information about a collection.')
        .modelParam('id', model='collection', level=AccessType.READ)
        .errorResponse()
        .errorResponse('Read access was denied on the collection.', 403)
    )
    def getCollectionDetails(self, collection, params):
        return {
            'nFolders': self.model('collection').countFolders(
                collection, user=self.getCurrentUser(), level=AccessType.READ)
        }

    @access.cookie
    @access.public(scope=TokenScope.DATA_READ)
    @autoDescribeRoute(
        Description('Download an entire collection as a zip archive.')
        .modelParam('id', model='collection', level=AccessType.READ)
        .jsonParam('mimeFilter', 'JSON list of MIME types to include.', requireArray=True,
                   required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Read access was denied for the collection.', 403)
    )
    def downloadCollection(self, collection, mimeFilter, params):
        setResponseHeader('Content-Type', 'application/zip')
        setResponseHeader(
            'Content-Disposition', 'attachment; filename="%s%s"' % (collection['name'], '.zip'))

        def stream():
            zip = ziputil.ZipGenerator(collection['name'])
            for (path, file) in self.model('collection').fileList(
                    collection, user=self.getCurrentUser(), subpath=False, mimeFilter=mimeFilter):
                for data in zip.addFile(file, path):
                    yield data
            yield zip.footer()
        return stream

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Get the access control list for a collection.')
        .modelParam('id', model='collection', level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def getCollectionAccess(self, collection, params):
        return self.model('collection').getFullAccessList(collection)

    @access.user(scope=TokenScope.DATA_OWN)
    @filtermodel(model='collection', addFields={'access'})
    @autoDescribeRoute(
        Description('Set the access control list for a collection.')
        .modelParam('id', model='collection', level=AccessType.ADMIN)
        .jsonParam('access', 'The access control list as JSON.', requireObject=True)
        .jsonParam('publicFlags', 'List of public access flags to set on the collection.',
                   required=False, requireArray=True)
        .param('public', 'Whether the collection should be publicly visible.',
               dataType='boolean', required=False)
        .param('recurse', 'Whether the policies should be applied to all '
               'folders under this collection as well.', dataType='boolean',
               default=False, required=False)
        .param('progress', 'If recurse is set to True, this controls whether '
               'progress notifications will be sent.', dataType='boolean',
               default=False, required=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def updateCollectionAccess(
            self, collection, access, public, recurse, progress, publicFlags, params):
        user = self.getCurrentUser()
        progress = progress and recurse

        with ProgressContext(progress, user=user, title='Updating permissions',
                             message='Calculating progress...') as ctx:
            if progress:
                ctx.update(total=self.model('collection').subtreeCount(
                    collection, includeItems=False, user=user,
                    level=AccessType.ADMIN))
            return self.model('collection').setAccessList(
                collection, access, save=True, user=user, recurse=recurse,
                progress=ctx, setPublic=public, publicFlags=publicFlags)

    @access.user(scope=TokenScope.DATA_READ)
    @filtermodel(model='collection')
    @autoDescribeRoute(
        Description('Edit a collection by ID.')
        .responseClass('Collection')
        .modelParam('id', model='collection', level=AccessType.WRITE)
        .param('name', 'Unique name for the collection.', required=False, strip=True)
        .param('description', 'Collection description.', required=False, strip=True)
        .errorResponse('ID was invalid.')
        .errorResponse('Write permission denied on the collection.', 403)
    )
    def updateCollection(self, collection, name, description, params):
        if name is not None:
            collection['name'] = name
        if description is not None:
            collection['description'] = description

        return self.model('collection').updateCollection(collection)

    @access.user(scope=TokenScope.DATA_OWN)
    @autoDescribeRoute(
        Description('Delete a collection by ID.')
        .modelParam('id', model='collection', level=AccessType.ADMIN)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin permission denied on the collection.', 403)
    )
    def deleteCollection(self, collection, params):
        self.model('collection').remove(collection)
        return {'message': 'Deleted collection %s.' % collection['name']}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  Copyright 2013 Kitware Inc.
#
#  Licensed under the Apache License, Version 2.0 ( the "License" );
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
###############################################################################

import datetime
import os
import re

from .model_base import AccessControlledModel, AccessException, ValidationException
from girder import events
from girder.constants import AccessType, CoreEventHandler, SettingKey, TokenScope
from girder.utility import config, mail_utils


class User(AccessControlledModel):
    """
    This model represents the users of the system.
    """

    def initialize(self):
        self.name = 'user'
        self.ensureIndices(['login', 'email', 'groupInvites.groupId'])
        self.prefixSearchFields = (
            'login', ('firstName', 'i'), ('lastName', 'i'))

        self.ensureTextIndex({
            'login': 1,
            'firstName': 1,
            'lastName': 1
        }, language='none')

        self.exposeFields(level=AccessType.READ, fields=(
            '_id', 'login', 'public', 'firstName', 'lastName', 'admin',
            'created'))
        self.exposeFields(level=AccessType.ADMIN, fields=(
            'size', 'email', 'groups', 'groupInvites', 'status',
            'emailVerified'))

        events.bind('model.user.save.created',
                    CoreEventHandler.USER_SELF_ACCESS, self._grantSelfAccess)
        events.bind('model.user.save.created',
                    CoreEventHandler.USER_DEFAULT_FOLDERS,
                    self._addDefaultFolders)

    def validate(self, doc):
        """
        Validate the user every time it is stored in the database.
        """
        doc['login'] = doc.get('login', '').lower().strip()
        doc['email'] = doc.get('email', '').lower().strip()
        doc['firstName'] = doc.get('firstName', '').strip()
        doc['lastName'] = doc.get('lastName', '').strip()
        doc['status'] = doc.get('status', 'enabled')

        cur_config = config.getConfig()

        if 'salt' not in doc:  # pragma: no cover
            # Internal error, this should not happen
            raise Exception('Tried to save user document with no salt.')

        if not doc['firstName']:
            raise ValidationException('First name must not be empty.',
                                      'firstName')

        if not doc['lastName']:
            raise ValidationException('Last name must not be empty.',
                                      'lastName')

        if doc['status'] not in ('pending', 'enabled', 'disabled'):
            raise ValidationException(
                'Status must be pending, enabled, or disabled.', 'status')

        if '@' in doc['login']:
            # Hard-code this constraint so we can always easily distinguish
            # an email address from a login
            raise ValidationException('Login may not contain "@".', 'login')

        if not re.match(cur_config['users']['login_regex'], doc['login']):
            raise ValidationException(
                cur_config['users']['login_description'], 'login')

        if not re.match(cur_config['users']['email_regex'], doc['email']):
            raise ValidationException('Invalid email address.', 'email')

        # Ensure unique logins
        q = {'login': doc['login']}
        if '_id' in doc:
            q['_id'] = {'$ne': doc['_id']}
        existing = self.findOne(q)
        if existing is not None:
            raise ValidationException('That login is already registered.',
                                      'login')

        # Ensure unique emails
        q = {'email': doc['email']}
        if '_id' in doc:
            q['_id'] = {'$ne': doc['_id']}
        existing = self.findOne(q)
        if existing is not None:
            raise ValidationException('That email is already registered.',
                                      'email')

        # If this is the first user being created, make it an admin
        existing = self.findOne({})
        if existing is None:
            doc['admin'] = True
            # Ensure settings don't stop this user from logging in
            doc['emailVerified'] = True
            doc['status'] = 'enabled'

        return doc

    def authenticate(self, login, password):
        """
        Validate a user login via username and password. If authentication fails,
        a ``AccessException`` is raised.

        :param login: The user's login or email.
        :type login: str
        :param password: The user's password.
        :type password: str
        :returns: The corresponding user if the login was successful.
        :rtype: dict
        """
        login = login.lower().strip()
        loginField = 'email' if '@' in login else 'login'

        user = self.model('user').findOne({loginField: login})
        if user is None:
            raise AccessException('Login failed.')

        if not self.model('password').authenticate(user, password):
            raise AccessException('Login failed.')

        # This has the same behavior as User.canLogin, but returns more
        # detailed error messages
        if user.get('status', 'enabled') == 'disabled':
            raise AccessException('Account is disabled.', extra='disabled')

        if self.model('user').emailVerificationRequired(user):
            raise AccessException(
                'Email verification required.', extra='emailVerification')

        if self.model('user').adminApprovalRequired(user):
            raise AccessException('Account approval required.', extra='accountApproval')

        return user

    def remove(self, user, progress=None, **kwargs):
        """
        Delete a user, and all references to it in the database.

        :param user: The user document to delete.
        :type user: dict
        :param progress: A progress context to record progress on.
        :type progress: girder.utility.progress.ProgressContext or None.
        """
        # Delete all authentication tokens owned by this user
        self.model('token').removeWithQuery({'userId': user['_id']})

        # Delete all pending group invites for this user
        self.model('group').update(
            {'requests': user['_id']},
            {'$pull': {'requests': user['_id']}}
        )

        # Delete all of the folders under this user
        folders = self.model('folder').find({
            'parentId': user['_id'],
            'parentCollection': 'user'
        })
        for folder in folders:
            self.model('folder').remove(folder, progress=progress, **kwargs)

        # Finally, delete the user document itself
        AccessControlledModel.remove(self, user)
        if progress:
            progress.update(increment=1, message='Deleted user ' +
                            user['login'])

    def getAdmins(self):
        """
        Helper to return a cursor of all site-admin users. The number of site
        admins is assumed to be small enough that we will not need to page the
        results for now.
        """
        return self.find({'admin': True})

    def search(self, text=None, user=None, limit=0, offset=0, sort=None):
        """
        List all users. Since users are access-controlled, this will filter
        them by access policy.

        :param text: Pass this to perform a full-text search for users.
        :param user: The user running the query. Only returns users that this
                     user can see.
        :param limit: Result limit.
        :param offset: Result offset.
        :param sort: The sort structure to pass to pymongo.
        :returns: Iterable of users.
        """
        # Perform the find; we'll do access-based filtering of the result set
        # afterward.
        if text is not None:
            cursor = self.textSearch(text, sort=sort)
        else:
            cursor = self.find({}, sort=sort)

        return self.filterResultsByPermission(
            cursor=cursor, user=user, level=AccessType.READ, limit=limit,
            offset=offset)

    def setPassword(self, user, password, save=True):
        """
        Change a user's password.

        :param user: The user whose password to change.
        :param password: The new password. If set to None, no password will
                         be stored for this user. This should be done in cases
                         where an external system is responsible for
                         authenticating the user.
        """
        if password is None:
            user['salt'] = None
        else:
            salt, alg = self.model('password').encryptAndStore(password)
            user['salt'] = salt
            user['hashAlg'] = alg

        if save:
            self.save(user)

    def createUser(self, login, password, firstName, lastName, email,
                   admin=False, public=True):
        """
        Create a new user with the given information. The user will be created
        with the default "Public" and "Private" folders.

        :param admin: Whether user is global administrator.
        :type admin: bool
        :param public: Whether user is publicly visible.
        :type public: bool
        :returns: The user document that was created.
        """
        requireApproval = self.model('setting').get(
            SettingKey.REGISTRATION_POLICY) == 'approve'
        if admin:
            requireApproval = False

        user = {
            'login': login,
            'email': email,
            'firstName': firstName,
            'lastName': lastName,
            'created': datetime.datetime.utcnow(),
            'emailVerified': False,
            'status': 'pending' if requireApproval else 'enabled',
            'admin': admin,
            'size': 0,
            'groups': [],
            'groupInvites': []
        }

        self.setPassword(user, password, save=False)
        self.setPublic(user, public, save=False)

        user = self.save(user)

        verifyEmail = self.model('setting').get(
            SettingKey.EMAIL_VERIFICATION) != 'disabled'
        if verifyEmail:
            self._sendVerificationEmail(user)

        if requireApproval:
            self._sendApprovalEmail(user)

        return user

    def canLogin(self, user):
        """
        Returns True if the user is allowed to login, e.g. email verification
        is not needed and admin approval is not needed.
        """
        if user.get('status', 'enabled') == 'disabled':
            return False
        if self.emailVerificationRequired(user):
            return False
        if self.adminApprovalRequired(user):
            return False
        return True

    def emailVerificationRequired(self, user):
        """
        Returns True if email verification is required and this user has not
        yet verified their email address.
        """
        return (not user['emailVerified']) and self.model('setting').get(
            SettingKey.EMAIL_VERIFICATION) == 'required'

    def adminApprovalRequired(self, user):
        """
        Returns True if the registration policy requires admin approval and
        this user is pending approval.
        """
        return user.get('status', 'enabled') == 'pending' and \
            self.model('setting').get(
                SettingKey.REGISTRATION_POLICY) == 'approve'

    def _sendApprovalEmail(self, user):
        url = '%s#user/%s' % (
            mail_utils.getEmailUrlPrefix(), str(user['_id']))
        text = mail_utils.renderTemplate('accountApproval.mako', {
            'user': user,
            'url': url
        })
        mail_utils.sendEmail(
            toAdmins=True,
            subject='Girder: Account pending approval',
            text=text)

    def _sendApprovedEmail(self, user):
        text = mail_utils.renderTemplate('accountApproved.mako', {
            'user': user,
            'url': mail_utils.getEmailUrlPrefix()
        })
        mail_utils.sendEmail(
            to=user.get('email'),
            subject='Girder: Account approved',
            text=text)

    def _sendVerificationEmail(self, user):
        token = self.model('token').createToken(
            user, days=1, scope=TokenScope.EMAIL_VERIFICATION)
        url = '%s#useraccount/%s/verification/%s' % (
            mail_utils.getEmailUrlPrefix(), str(user['_id']), str(token['_id']))
        text = mail_utils.renderTemplate('emailVerification.mako', {
            'url': url
        })
        mail_utils.sendEmail(
            to=user.get('email'),
            subject='Girder: Email verification',
            text=text)

    def _grantSelfAccess(self, event):
        """
        This callback grants a user admin access to itself.

        This generally should not be called or overridden directly, but it may
        be unregistered from the `model.user.save.created` event.
        """
        user = event.info

        self.setUserAccess(user, user, level=AccessType.ADMIN, save=True)

    def _addDefaultFolders(self, event):
        """
        This callback creates "Public" and "Private" folders on a user, after
        it is first created.

        This generally should not be called or overridden directly, but it may
        be unregistered from the `model.user.save.created` event.
        """
        if self.model('setting').get(
                SettingKey.USER_DEFAULT_FOLDERS, 'public_private') \
                == 'public_private':
            user = event.info

            publicFolder = self.model('folder').createFolder(
                user, 'Public', parentType='user', public=True, creator=user)
            privateFolder = self.model('folder').createFolder(
                user, 'Private', parentType='user', public=False, creator=user)
            # Give the user admin access to their own folders
            self.model('folder').setUserAccess(
                publicFolder, user, AccessType.ADMIN, save=True)
            self.model('folder').setUserAccess(
                privateFolder, user, AccessType.ADMIN, save=True)

    def fileList(self, doc, user=None, path='', includeMetadata=False,
                 subpath=True, data=True):
        """
        This function generates a list of 2-tuples whose first element is the
        relative path to the file from the user's folders root and whose second
        element depends on the value of the `data` flag. If `data=True`, the
        second element will be a generator that will generate the bytes of the
        file data as stored in the assetstore. If `data=False`, the second
        element is the file document itself.

        :param doc: the user to list.
        :param user: a user used to validate data that is returned.
        :param path: a path prefix to add to the results.
        :param includeMetadata: if True and there is any metadata, include a
                                result which is the JSON string of the
                                metadata.  This is given a name of
                                metadata[-(number).json that is distinct from
                                any file within the item.
        :param subpath: if True, add the user's name to the path.
        :param data: If True return raw content of each file as stored in the
            assetstore, otherwise return file document.
        :type data: bool
        """
        if subpath:
            path = os.path.join(path, doc['login'])
        for folder in self.model('folder').childFolders(parentType='user',
                                                        parent=doc, user=user):
            for (filepath, file) in self.model('folder').fileList(
                    folder, user, path, includeMetadata, subpath=True,
                    data=data):
                yield (filepath, file)

    def subtreeCount(self, doc, includeItems=True, user=None, level=None):
        """
        Return the size of the user's folders.  The user is counted as well.

        :param doc: The user.
        :param includeItems: Whether to include items in the subtree count, or
            just folders.
        :type includeItems: bool
        :param user: If filtering by permission, the user to filter against.
        :param level: If filtering by permission, the required permission level.
        :type level: AccessLevel
        """
        count = 1
        folders = self.model('folder').find({
            'parentId': doc['_id'],
            'parentCollection': 'user'
        }, fields=('access',))

        if level is not None:
            folders = self.filterResultsByPermission(
                cursor=folders, user=user, level=level)

        count += sum(self.model('folder').subtreeCount(
            folder, includeItems=includeItems, user=user, level=level)
            for folder in folders)
        return count

    def countFolders(self, user, filterUser=None, level=None):
        """
        Returns the number of top level folders under this user. Access
        checking is optional; to circumvent access checks, pass ``level=None``.

        :param user: The user whose top level folders to count.
        :type collection: dict
        :param filterUser: If performing access checks, the user to check
            against.
        :type filterUser: dict or None
        :param level: The required access level, or None to return the raw
            top-level folder count.
        """
        fields = () if level is None else ('access', 'public')

        folderModel = self.model('folder')
        folders = folderModel.find({
            'parentId': user['_id'],
            'parentCollection': 'user'
        }, fields=fields)

        if level is None:
            return folders.count()
        else:
            return sum(1 for _ in folderModel.filterResultsByPermission(
                cursor=folders, user=filterUser, level=level))

    def updateSize(self, doc):
        """
        Recursively recomputes the size of this user and its underlying
        folders and fixes the sizes as needed.

        :param doc: The user.
        :type doc: dict
        """
        size = 0
        fixes = 0
        folders = self.model('folder').find({
            'parentId': doc['_id'],
            'parentCollection': 'user'
        })
        for folder in folders:
            # fix folder size if needed
            _, f = self.model('folder').updateSize(folder)
            fixes += f
            # get total recursive folder size
            folder = self.model('folder').load(folder['_id'], force=True)
            size += self.model('folder').getSizeRecursive(folder)
        # fix value if incorrect
        if size != doc.get('size'):
            self.update({'_id': doc['_id']}, update={'$set': {'size': size}})
            fixes += 1
        return size, fixes
